#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This package is deprecated. Please use `airflow.operators` or `airflow.providers.*.operators`."""
from __future__ import annotations

import warnings

from airflow.exceptions import RemovedInAirflow3Warning
from airflow.utils.deprecation_tools import add_deprecated_classes

warnings.warn(
    "This package is deprecated. Please use `airflow.operators` or `airflow.providers.*.operators`.",
    RemovedInAirflow3Warning,
    stacklevel=2,
)

__deprecated_classes = {
    'adls_list_operator': {
        'ADLSListOperator': 'airflow.providers.microsoft.azure.operators.adls.ADLSListOperator',
        'AzureDataLakeStorageListOperator': (
            'airflow.providers.microsoft.azure.operators.adls.ADLSListOperator'
        ),
    },
    'adls_to_gcs': {
        'ADLSToGCSOperator': 'airflow.providers.google.cloud.transfers.adls_to_gcs.ADLSToGCSOperator',
        'AdlsToGoogleCloudStorageOperator': (
            'airflow.providers.google.cloud.transfers.adls_to_gcs.ADLSToGCSOperator'
        ),
    },
    'aws_athena_operator': {
        'AWSAthenaOperator': 'airflow.providers.amazon.aws.operators.athena.AthenaOperator',
    },
    'aws_sqs_publish_operator': {
        'SqsPublishOperator': 'airflow.providers.amazon.aws.operators.sqs.SqsPublishOperator',
        'SQSPublishOperator': 'airflow.providers.amazon.aws.operators.sqs.SqsPublishOperator',
    },
    'awsbatch_operator': {
        'BatchProtocol': 'airflow.providers.amazon.aws.hooks.batch_client.BatchProtocol',
        'BatchOperator': 'airflow.providers.amazon.aws.operators.batch.BatchOperator',
        'AWSBatchOperator': 'airflow.providers.amazon.aws.operators.batch.BatchOperator',
    },
    'azure_container_instances_operator': {
        'AzureContainerInstancesOperator': (
            'airflow.providers.microsoft.azure.operators.container_instances.AzureContainerInstancesOperator'
        ),
    },
    'azure_cosmos_operator': {
        'AzureCosmosInsertDocumentOperator': (
            'airflow.providers.microsoft.azure.operators.cosmos.AzureCosmosInsertDocumentOperator'
        ),
    },
    'bigquery_check_operator': {
        'BigQueryCheckOperator': 'airflow.providers.google.cloud.operators.bigquery.BigQueryCheckOperator',
        'BigQueryIntervalCheckOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryIntervalCheckOperator'
        ),
        'BigQueryValueCheckOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryValueCheckOperator'
        ),
    },
    'bigquery_get_data': {
        'BigQueryGetDataOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryGetDataOperator'
        ),
    },
    'bigquery_operator': {
        'BigQueryCreateEmptyDatasetOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryCreateEmptyDatasetOperator'
        ),
        'BigQueryCreateEmptyTableOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryCreateEmptyTableOperator'
        ),
        'BigQueryCreateExternalTableOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryCreateExternalTableOperator'
        ),
        'BigQueryDeleteDatasetOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryDeleteDatasetOperator'
        ),
        'BigQueryExecuteQueryOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryExecuteQueryOperator'
        ),
        'BigQueryGetDatasetOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryGetDatasetOperator'
        ),
        'BigQueryGetDatasetTablesOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryGetDatasetTablesOperator'
        ),
        'BigQueryPatchDatasetOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryPatchDatasetOperator'
        ),
        'BigQueryUpdateDatasetOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryUpdateDatasetOperator'
        ),
        'BigQueryUpsertTableOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryUpsertTableOperator'
        ),
        'BigQueryOperator': 'airflow.providers.google.cloud.operators.bigquery.BigQueryExecuteQueryOperator',
    },
    'bigquery_table_delete_operator': {
        'BigQueryDeleteTableOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryDeleteTableOperator'
        ),
        'BigQueryTableDeleteOperator': (
            'airflow.providers.google.cloud.operators.bigquery.BigQueryDeleteTableOperator'
        ),
    },
    'bigquery_to_bigquery': {
        'BigQueryToBigQueryOperator': (
            'airflow.providers.google.cloud.transfers.bigquery_to_bigquery.BigQueryToBigQueryOperator'
        ),
    },
    'bigquery_to_gcs': {
        'BigQueryToGCSOperator': (
            'airflow.providers.google.cloud.transfers.bigquery_to_gcs.BigQueryToGCSOperator'
        ),
        'BigQueryToCloudStorageOperator': (
            'airflow.providers.google.cloud.transfers.bigquery_to_gcs.BigQueryToGCSOperator'
        ),
    },
    'bigquery_to_mysql_operator': {
        'BigQueryToMySqlOperator': (
            'airflow.providers.google.cloud.transfers.bigquery_to_mysql.BigQueryToMySqlOperator'
        ),
    },
    'cassandra_to_gcs': {
        'CassandraToGCSOperator': (
            'airflow.providers.google.cloud.transfers.cassandra_to_gcs.CassandraToGCSOperator'
        ),
        'CassandraToGoogleCloudStorageOperator': (
            'airflow.providers.google.cloud.transfers.cassandra_to_gcs.CassandraToGCSOperator'
        ),
    },
    'databricks_operator': {
        'DatabricksRunNowOperator': (
            'airflow.providers.databricks.operators.databricks.DatabricksRunNowOperator'
        ),
        'DatabricksSubmitRunOperator': (
            'airflow.providers.databricks.operators.databricks.DatabricksSubmitRunOperator'
        ),
    },
    'dataflow_operator': {
        'DataflowCreateJavaJobOperator': (
            'airflow.providers.google.cloud.operators.dataflow.DataflowCreateJavaJobOperator'
        ),
        'DataflowCreatePythonJobOperator': (
            'airflow.providers.google.cloud.operators.dataflow.DataflowCreatePythonJobOperator'
        ),
        'DataflowTemplatedJobStartOperator': (
            'airflow.providers.google.cloud.operators.dataflow.DataflowTemplatedJobStartOperator'
        ),
        'DataFlowJavaOperator': (
            'airflow.providers.google.cloud.operators.dataflow.DataflowCreateJavaJobOperator'
        ),
        'DataFlowPythonOperator': (
            'airflow.providers.google.cloud.operators.dataflow.DataflowCreatePythonJobOperator'
        ),
        'DataflowTemplateOperator': (
            'airflow.providers.google.cloud.operators.dataflow.DataflowTemplatedJobStartOperator'
        ),
    },
    'dataproc_operator': {
        'DataprocCreateClusterOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocCreateClusterOperator'
        ),
        'DataprocDeleteClusterOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocDeleteClusterOperator'
        ),
        'DataprocInstantiateInlineWorkflowTemplateOperator':
            'airflow.providers.google.cloud.operators.dataproc.'
            'DataprocInstantiateInlineWorkflowTemplateOperator',
        'DataprocInstantiateWorkflowTemplateOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocInstantiateWorkflowTemplateOperator'
        ),
        'DataprocJobBaseOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocJobBaseOperator'
        ),
        'DataprocScaleClusterOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocScaleClusterOperator'
        ),
        'DataprocSubmitHadoopJobOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocSubmitHadoopJobOperator'
        ),
        'DataprocSubmitHiveJobOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocSubmitHiveJobOperator'
        ),
        'DataprocSubmitPigJobOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocSubmitPigJobOperator'
        ),
        'DataprocSubmitPySparkJobOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocSubmitPySparkJobOperator'
        ),
        'DataprocSubmitSparkJobOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocSubmitSparkJobOperator'
        ),
        'DataprocSubmitSparkSqlJobOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocSubmitSparkSqlJobOperator'
        ),
        'DataprocClusterCreateOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocCreateClusterOperator'
        ),
        'DataprocClusterDeleteOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocDeleteClusterOperator'
        ),
        'DataprocClusterScaleOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocScaleClusterOperator'
        ),
        'DataProcHadoopOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocSubmitHadoopJobOperator'
        ),
        'DataProcHiveOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocSubmitHiveJobOperator'
        ),
        'DataProcJobBaseOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocJobBaseOperator'
        ),
        'DataProcPigOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocSubmitPigJobOperator'
        ),
        'DataProcPySparkOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocSubmitPySparkJobOperator'
        ),
        'DataProcSparkOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocSubmitSparkJobOperator'
        ),
        'DataProcSparkSqlOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocSubmitSparkSqlJobOperator'
        ),
        'DataprocWorkflowTemplateInstantiateInlineOperator':
            'airflow.providers.google.cloud.operators.dataproc.'
            'DataprocInstantiateInlineWorkflowTemplateOperator',
        'DataprocWorkflowTemplateInstantiateOperator': (
            'airflow.providers.google.cloud.operators.dataproc.DataprocInstantiateWorkflowTemplateOperator'
        ),
    },
    'datastore_export_operator': {
        'CloudDatastoreExportEntitiesOperator': (
            'airflow.providers.google.cloud.operators.datastore.CloudDatastoreExportEntitiesOperator'
        ),
        'DatastoreExportOperator': (
            'airflow.providers.google.cloud.operators.datastore.CloudDatastoreExportEntitiesOperator'
        ),
    },
    'datastore_import_operator': {
        'CloudDatastoreImportEntitiesOperator': (
            'airflow.providers.google.cloud.operators.datastore.CloudDatastoreImportEntitiesOperator'
        ),
        'DatastoreImportOperator': (
            'airflow.providers.google.cloud.operators.datastore.CloudDatastoreImportEntitiesOperator'
        ),
    },
    'dingding_operator': {
        'DingdingOperator': 'airflow.providers.dingding.operators.dingding.DingdingOperator',
    },
    'discord_webhook_operator': {
        'DiscordWebhookOperator': (
            'airflow.providers.discord.operators.discord_webhook.DiscordWebhookOperator'
        ),
    },
    'docker_swarm_operator': {
        'DockerSwarmOperator': 'airflow.providers.docker.operators.docker_swarm.DockerSwarmOperator',
    },
    'druid_operator': {
        'DruidOperator': 'airflow.providers.apache.druid.operators.druid.DruidOperator',
    },
    'dynamodb_to_s3': {
        'DynamoDBToS3Operator': 'airflow.providers.amazon.aws.transfers.dynamodb_to_s3.DynamoDBToS3Operator',
    },
    'ecs_operator': {
        'EcsProtocol': 'airflow.providers.amazon.aws.hooks.ecs.EcsProtocol',
        'EcsRunTaskOperator': 'airflow.providers.amazon.aws.operators.ecs.EcsRunTaskOperator',
        'EcsOperator': 'airflow.providers.amazon.aws.operators.ecs.EcsRunTaskOperator',
    },
    'file_to_gcs': {
        'LocalFilesystemToGCSOperator': (
            'airflow.providers.google.cloud.transfers.local_to_gcs.LocalFilesystemToGCSOperator'
        ),
        'FileToGoogleCloudStorageOperator': (
            'airflow.providers.google.cloud.transfers.local_to_gcs.LocalFilesystemToGCSOperator'
        ),
    },
    'file_to_wasb': {
        'LocalFilesystemToWasbOperator': (
            'airflow.providers.microsoft.azure.transfers.local_to_wasb.LocalFilesystemToWasbOperator'
        ),
        'FileToWasbOperator': (
            'airflow.providers.microsoft.azure.transfers.local_to_wasb.LocalFilesystemToWasbOperator'
        ),
    },
    'gcp_bigtable_operator': {
        'BigtableCreateInstanceOperator': (
            'airflow.providers.google.cloud.operators.bigtable.BigtableCreateInstanceOperator'
        ),
        'BigtableCreateTableOperator': (
            'airflow.providers.google.cloud.operators.bigtable.BigtableCreateTableOperator'
        ),
        'BigtableDeleteInstanceOperator': (
            'airflow.providers.google.cloud.operators.bigtable.BigtableDeleteInstanceOperator'
        ),
        'BigtableDeleteTableOperator': (
            'airflow.providers.google.cloud.operators.bigtable.BigtableDeleteTableOperator'
        ),
        'BigtableUpdateClusterOperator': (
            'airflow.providers.google.cloud.operators.bigtable.BigtableUpdateClusterOperator'
        ),
        'BigtableTableReplicationCompletedSensor': (
            'airflow.providers.google.cloud.sensors.bigtable.BigtableTableReplicationCompletedSensor'
        ),
        'BigtableClusterUpdateOperator': (
            'airflow.providers.google.cloud.operators.bigtable.BigtableUpdateClusterOperator'
        ),
        'BigtableInstanceCreateOperator': (
            'airflow.providers.google.cloud.operators.bigtable.BigtableCreateInstanceOperator'
        ),
        'BigtableInstanceDeleteOperator': (
            'airflow.providers.google.cloud.operators.bigtable.BigtableDeleteInstanceOperator'
        ),
        'BigtableTableCreateOperator': (
            'airflow.providers.google.cloud.operators.bigtable.BigtableCreateTableOperator'
        ),
        'BigtableTableDeleteOperator': (
            'airflow.providers.google.cloud.operators.bigtable.BigtableDeleteTableOperator'
        ),
        'BigtableTableWaitForReplicationSensor': (
            'airflow.providers.google.cloud.sensors.bigtable.BigtableTableReplicationCompletedSensor'
        ),
    },
    'gcp_cloud_build_operator': {
        'CloudBuildCreateBuildOperator': (
            'airflow.providers.google.cloud.operators.cloud_build.CloudBuildCreateBuildOperator'
        ),
    },
    'gcp_compute_operator': {
        'ComputeEngineBaseOperator': (
            'airflow.providers.google.cloud.operators.compute.ComputeEngineBaseOperator'
        ),
        'ComputeEngineCopyInstanceTemplateOperator': (
            'airflow.providers.google.cloud.operators.compute.ComputeEngineCopyInstanceTemplateOperator'
        ),
        'ComputeEngineInstanceGroupUpdateManagerTemplateOperator':
            'airflow.providers.google.cloud.operators.compute.'
            'ComputeEngineInstanceGroupUpdateManagerTemplateOperator',
        'ComputeEngineSetMachineTypeOperator': (
            'airflow.providers.google.cloud.operators.compute.ComputeEngineSetMachineTypeOperator'
        ),
        'ComputeEngineStartInstanceOperator': (
            'airflow.providers.google.cloud.operators.compute.ComputeEngineStartInstanceOperator'
        ),
        'ComputeEngineStopInstanceOperator': (
            'airflow.providers.google.cloud.operators.compute.ComputeEngineStopInstanceOperator'
        ),
        'GceBaseOperator': 'airflow.providers.google.cloud.operators.compute.ComputeEngineBaseOperator',
        'GceInstanceGroupManagerUpdateTemplateOperator':
            'airflow.providers.google.cloud.operators.compute.'
            'ComputeEngineInstanceGroupUpdateManagerTemplateOperator',
        'GceInstanceStartOperator': (
            'airflow.providers.google.cloud.operators.compute.ComputeEngineStartInstanceOperator'
        ),
        'GceInstanceStopOperator': (
            'airflow.providers.google.cloud.operators.compute.ComputeEngineStopInstanceOperator'
        ),
        'GceInstanceTemplateCopyOperator': (
            'airflow.providers.google.cloud.operators.compute.ComputeEngineCopyInstanceTemplateOperator'
        ),
        'GceSetMachineTypeOperator': (
            'airflow.providers.google.cloud.operators.compute.ComputeEngineSetMachineTypeOperator'
        ),
    },
    'gcp_container_operator': {
        'GKECreateClusterOperator': (
            'airflow.providers.google.cloud.operators.kubernetes_engine.GKECreateClusterOperator'
        ),
        'GKEDeleteClusterOperator': (
            'airflow.providers.google.cloud.operators.kubernetes_engine.GKEDeleteClusterOperator'
        ),
        'GKEStartPodOperator': (
            'airflow.providers.google.cloud.operators.kubernetes_engine.GKEStartPodOperator'
        ),
        'GKEClusterCreateOperator': (
            'airflow.providers.google.cloud.operators.kubernetes_engine.GKECreateClusterOperator'
        ),
        'GKEClusterDeleteOperator': (
            'airflow.providers.google.cloud.operators.kubernetes_engine.GKEDeleteClusterOperator'
        ),
        'GKEPodOperator': 'airflow.providers.google.cloud.operators.kubernetes_engine.GKEStartPodOperator',
    },
    'gcp_dlp_operator': {
        'CloudDLPCancelDLPJobOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPCancelDLPJobOperator'
        ),
        'CloudDLPCreateDeidentifyTemplateOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPCreateDeidentifyTemplateOperator'
        ),
        'CloudDLPCreateDLPJobOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPCreateDLPJobOperator'
        ),
        'CloudDLPCreateInspectTemplateOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPCreateInspectTemplateOperator'
        ),
        'CloudDLPCreateJobTriggerOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPCreateJobTriggerOperator'
        ),
        'CloudDLPCreateStoredInfoTypeOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPCreateStoredInfoTypeOperator'
        ),
        'CloudDLPDeidentifyContentOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPDeidentifyContentOperator'
        ),
        'CloudDLPDeleteDeidentifyTemplateOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPDeleteDeidentifyTemplateOperator'
        ),
        'CloudDLPDeleteDLPJobOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPDeleteDLPJobOperator'
        ),
        'CloudDLPDeleteInspectTemplateOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPDeleteInspectTemplateOperator'
        ),
        'CloudDLPDeleteJobTriggerOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPDeleteJobTriggerOperator'
        ),
        'CloudDLPDeleteStoredInfoTypeOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPDeleteStoredInfoTypeOperator'
        ),
        'CloudDLPGetDeidentifyTemplateOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPGetDeidentifyTemplateOperator'
        ),
        'CloudDLPGetDLPJobOperator': 'airflow.providers.google.cloud.operators.dlp.CloudDLPGetDLPJobOperator',
        'CloudDLPGetDLPJobTriggerOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPGetDLPJobTriggerOperator'
        ),
        'CloudDLPGetInspectTemplateOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPGetInspectTemplateOperator'
        ),
        'CloudDLPGetStoredInfoTypeOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPGetStoredInfoTypeOperator'
        ),
        'CloudDLPInspectContentOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPInspectContentOperator'
        ),
        'CloudDLPListDeidentifyTemplatesOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPListDeidentifyTemplatesOperator'
        ),
        'CloudDLPListDLPJobsOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPListDLPJobsOperator'
        ),
        'CloudDLPListInfoTypesOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPListInfoTypesOperator'
        ),
        'CloudDLPListInspectTemplatesOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPListInspectTemplatesOperator'
        ),
        'CloudDLPListJobTriggersOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPListJobTriggersOperator'
        ),
        'CloudDLPListStoredInfoTypesOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPListStoredInfoTypesOperator'
        ),
        'CloudDLPRedactImageOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPRedactImageOperator'
        ),
        'CloudDLPReidentifyContentOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPReidentifyContentOperator'
        ),
        'CloudDLPUpdateDeidentifyTemplateOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPUpdateDeidentifyTemplateOperator'
        ),
        'CloudDLPUpdateInspectTemplateOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPUpdateInspectTemplateOperator'
        ),
        'CloudDLPUpdateJobTriggerOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPUpdateJobTriggerOperator'
        ),
        'CloudDLPUpdateStoredInfoTypeOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPUpdateStoredInfoTypeOperator'
        ),
        'CloudDLPDeleteDlpJobOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPDeleteDLPJobOperator'
        ),
        'CloudDLPGetDlpJobOperator': 'airflow.providers.google.cloud.operators.dlp.CloudDLPGetDLPJobOperator',
        'CloudDLPGetJobTripperOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPGetDLPJobTriggerOperator'
        ),
        'CloudDLPListDlpJobsOperator': (
            'airflow.providers.google.cloud.operators.dlp.CloudDLPListDLPJobsOperator'
        ),
    },
    'gcp_function_operator': {
        'CloudFunctionDeleteFunctionOperator': (
            'airflow.providers.google.cloud.operators.functions.CloudFunctionDeleteFunctionOperator'
        ),
        'CloudFunctionDeployFunctionOperator': (
            'airflow.providers.google.cloud.operators.functions.CloudFunctionDeployFunctionOperator'
        ),
        'GcfFunctionDeleteOperator': (
            'airflow.providers.google.cloud.operators.functions.CloudFunctionDeleteFunctionOperator'
        ),
        'GcfFunctionDeployOperator': (
            'airflow.providers.google.cloud.operators.functions.CloudFunctionDeployFunctionOperator'
        ),
    },
    'gcp_natural_language_operator': {
        'CloudNaturalLanguageAnalyzeEntitiesOperator':
            'airflow.providers.google.cloud.operators.natural_language.'
            'CloudNaturalLanguageAnalyzeEntitiesOperator',
        'CloudNaturalLanguageAnalyzeEntitySentimentOperator':
            'airflow.providers.google.cloud.operators.natural_language.'
            'CloudNaturalLanguageAnalyzeEntitySentimentOperator',
        'CloudNaturalLanguageAnalyzeSentimentOperator':
            'airflow.providers.google.cloud.operators.natural_language.'
            'CloudNaturalLanguageAnalyzeSentimentOperator',
        'CloudNaturalLanguageClassifyTextOperator':
            'airflow.providers.google.cloud.operators.natural_language.'
            'CloudNaturalLanguageClassifyTextOperator',
        'CloudLanguageAnalyzeEntitiesOperator':
            'airflow.providers.google.cloud.operators.natural_language.'
            'CloudNaturalLanguageAnalyzeEntitiesOperator',
        'CloudLanguageAnalyzeEntitySentimentOperator':
            'airflow.providers.google.cloud.operators.natural_language.'
            'CloudNaturalLanguageAnalyzeEntitySentimentOperator',
        'CloudLanguageAnalyzeSentimentOperator':
            'airflow.providers.google.cloud.operators.natural_language.'
            'CloudNaturalLanguageAnalyzeSentimentOperator',
        'CloudLanguageClassifyTextOperator':
            'airflow.providers.google.cloud.operators.natural_language.'
            'CloudNaturalLanguageClassifyTextOperator',
    },
    'gcp_spanner_operator': {
        'SpannerDeleteDatabaseInstanceOperator': (
            'airflow.providers.google.cloud.operators.spanner.SpannerDeleteDatabaseInstanceOperator'
        ),
        'SpannerDeleteInstanceOperator': (
            'airflow.providers.google.cloud.operators.spanner.SpannerDeleteInstanceOperator'
        ),
        'SpannerDeployDatabaseInstanceOperator': (
            'airflow.providers.google.cloud.operators.spanner.SpannerDeployDatabaseInstanceOperator'
        ),
        'SpannerDeployInstanceOperator': (
            'airflow.providers.google.cloud.operators.spanner.SpannerDeployInstanceOperator'
        ),
        'SpannerQueryDatabaseInstanceOperator': (
            'airflow.providers.google.cloud.operators.spanner.SpannerQueryDatabaseInstanceOperator'
        ),
        'SpannerUpdateDatabaseInstanceOperator': (
            'airflow.providers.google.cloud.operators.spanner.SpannerUpdateDatabaseInstanceOperator'
        ),
        'CloudSpannerInstanceDatabaseDeleteOperator': (
            'airflow.providers.google.cloud.operators.spanner.SpannerDeleteDatabaseInstanceOperator'
        ),
        'CloudSpannerInstanceDatabaseDeployOperator': (
            'airflow.providers.google.cloud.operators.spanner.SpannerDeployDatabaseInstanceOperator'
        ),
        'CloudSpannerInstanceDatabaseQueryOperator': (
            'airflow.providers.google.cloud.operators.spanner.SpannerQueryDatabaseInstanceOperator'
        ),
        'CloudSpannerInstanceDatabaseUpdateOperator': (
            'airflow.providers.google.cloud.operators.spanner.SpannerUpdateDatabaseInstanceOperator'
        ),
        'CloudSpannerInstanceDeleteOperator': (
            'airflow.providers.google.cloud.operators.spanner.SpannerDeleteInstanceOperator'
        ),
        'CloudSpannerInstanceDeployOperator': (
            'airflow.providers.google.cloud.operators.spanner.SpannerDeployInstanceOperator'
        ),
    },
    'gcp_speech_to_text_operator': {
        'CloudSpeechToTextRecognizeSpeechOperator': (
            'airflow.providers.google.cloud.operators.speech_to_text.CloudSpeechToTextRecognizeSpeechOperator'
        ),
        'GcpSpeechToTextRecognizeSpeechOperator': (
            'airflow.providers.google.cloud.operators.speech_to_text.CloudSpeechToTextRecognizeSpeechOperator'
        ),
    },
    'gcp_sql_operator': {
        'CloudSQLBaseOperator': 'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLBaseOperator',
        'CloudSQLCreateInstanceDatabaseOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLCreateInstanceDatabaseOperator'
        ),
        'CloudSQLCreateInstanceOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLCreateInstanceOperator'
        ),
        'CloudSQLDeleteInstanceDatabaseOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLDeleteInstanceDatabaseOperator'
        ),
        'CloudSQLDeleteInstanceOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLDeleteInstanceOperator'
        ),
        'CloudSQLExecuteQueryOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLExecuteQueryOperator'
        ),
        'CloudSQLExportInstanceOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLExportInstanceOperator'
        ),
        'CloudSQLImportInstanceOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLImportInstanceOperator'
        ),
        'CloudSQLInstancePatchOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLInstancePatchOperator'
        ),
        'CloudSQLPatchInstanceDatabaseOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLPatchInstanceDatabaseOperator'
        ),
        'CloudSqlBaseOperator': 'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLBaseOperator',
        'CloudSqlInstanceCreateOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLCreateInstanceOperator'
        ),
        'CloudSqlInstanceDatabaseCreateOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLCreateInstanceDatabaseOperator'
        ),
        'CloudSqlInstanceDatabaseDeleteOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLDeleteInstanceDatabaseOperator'
        ),
        'CloudSqlInstanceDatabasePatchOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLPatchInstanceDatabaseOperator'
        ),
        'CloudSqlInstanceDeleteOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLDeleteInstanceOperator'
        ),
        'CloudSqlInstanceExportOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLExportInstanceOperator'
        ),
        'CloudSqlInstanceImportOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLImportInstanceOperator'
        ),
        'CloudSqlInstancePatchOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLInstancePatchOperator'
        ),
        'CloudSqlQueryOperator': (
            'airflow.providers.google.cloud.operators.cloud_sql.CloudSQLExecuteQueryOperator'
        ),
    },
    'gcp_tasks_operator': {
        'CloudTasksQueueCreateOperator': (
            'airflow.providers.google.cloud.operators.tasks.CloudTasksQueueCreateOperator'
        ),
        'CloudTasksQueueDeleteOperator': (
            'airflow.providers.google.cloud.operators.tasks.CloudTasksQueueDeleteOperator'
        ),
        'CloudTasksQueueGetOperator': (
            'airflow.providers.google.cloud.operators.tasks.CloudTasksQueueGetOperator'
        ),
        'CloudTasksQueuePauseOperator': (
            'airflow.providers.google.cloud.operators.tasks.CloudTasksQueuePauseOperator'
        ),
        'CloudTasksQueuePurgeOperator': (
            'airflow.providers.google.cloud.operators.tasks.CloudTasksQueuePurgeOperator'
        ),
        'CloudTasksQueueResumeOperator': (
            'airflow.providers.google.cloud.operators.tasks.CloudTasksQueueResumeOperator'
        ),
        'CloudTasksQueuesListOperator': (
            'airflow.providers.google.cloud.operators.tasks.CloudTasksQueuesListOperator'
        ),
        'CloudTasksQueueUpdateOperator': (
            'airflow.providers.google.cloud.operators.tasks.CloudTasksQueueUpdateOperator'
        ),
        'CloudTasksTaskCreateOperator': (
            'airflow.providers.google.cloud.operators.tasks.CloudTasksTaskCreateOperator'
        ),
        'CloudTasksTaskDeleteOperator': (
            'airflow.providers.google.cloud.operators.tasks.CloudTasksTaskDeleteOperator'
        ),
        'CloudTasksTaskGetOperator': (
            'airflow.providers.google.cloud.operators.tasks.CloudTasksTaskGetOperator'
        ),
        'CloudTasksTaskRunOperator': (
            'airflow.providers.google.cloud.operators.tasks.CloudTasksTaskRunOperator'
        ),
        'CloudTasksTasksListOperator': (
            'airflow.providers.google.cloud.operators.tasks.CloudTasksTasksListOperator'
        ),
    },
    'gcp_text_to_speech_operator': {
        'CloudTextToSpeechSynthesizeOperator': (
            'airflow.providers.google.cloud.operators.text_to_speech.CloudTextToSpeechSynthesizeOperator'
        ),
        'GcpTextToSpeechSynthesizeOperator': (
            'airflow.providers.google.cloud.operators.text_to_speech.CloudTextToSpeechSynthesizeOperator'
        ),
    },
    'gcp_transfer_operator': {
        'CloudDataTransferServiceCancelOperationOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceCancelOperationOperator',
        'CloudDataTransferServiceCreateJobOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceCreateJobOperator',
        'CloudDataTransferServiceDeleteJobOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceDeleteJobOperator',
        'CloudDataTransferServiceGCSToGCSOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceGCSToGCSOperator',
        'CloudDataTransferServiceGetOperationOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceGetOperationOperator',
        'CloudDataTransferServiceListOperationsOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceListOperationsOperator',
        'CloudDataTransferServicePauseOperationOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServicePauseOperationOperator',
        'CloudDataTransferServiceResumeOperationOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceResumeOperationOperator',
        'CloudDataTransferServiceS3ToGCSOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceS3ToGCSOperator',
        'CloudDataTransferServiceUpdateJobOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceUpdateJobOperator',
        'GcpTransferServiceJobCreateOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceCreateJobOperator',
        'GcpTransferServiceJobDeleteOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceDeleteJobOperator',
        'GcpTransferServiceJobUpdateOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceUpdateJobOperator',
        'GcpTransferServiceOperationCancelOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceCancelOperationOperator',
        'GcpTransferServiceOperationGetOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceGetOperationOperator',
        'GcpTransferServiceOperationPauseOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServicePauseOperationOperator',
        'GcpTransferServiceOperationResumeOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceResumeOperationOperator',
        'GcpTransferServiceOperationsListOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceListOperationsOperator',
        'GoogleCloudStorageToGoogleCloudStorageTransferOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceGCSToGCSOperator',
        'S3ToGoogleCloudStorageTransferOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceS3ToGCSOperator',
    },
    'gcp_translate_operator': {
        'CloudTranslateTextOperator': (
            'airflow.providers.google.cloud.operators.translate.CloudTranslateTextOperator'
        ),
    },
    'gcp_translate_speech_operator': {
        'CloudTranslateSpeechOperator': (
            'airflow.providers.google.cloud.operators.translate_speech.CloudTranslateSpeechOperator'
        ),
        'GcpTranslateSpeechOperator': (
            'airflow.providers.google.cloud.operators.translate_speech.CloudTranslateSpeechOperator'
        ),
    },
    'gcp_video_intelligence_operator': {
        'CloudVideoIntelligenceDetectVideoExplicitContentOperator':
            'airflow.providers.google.cloud.operators.video_intelligence.'
            'CloudVideoIntelligenceDetectVideoExplicitContentOperator',
        'CloudVideoIntelligenceDetectVideoLabelsOperator':
            'airflow.providers.google.cloud.operators.video_intelligence.'
            'CloudVideoIntelligenceDetectVideoLabelsOperator',
        'CloudVideoIntelligenceDetectVideoShotsOperator':
            'airflow.providers.google.cloud.operators.video_intelligence.'
            'CloudVideoIntelligenceDetectVideoShotsOperator',
    },
    'gcp_vision_operator': {
        'CloudVisionAddProductToProductSetOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionAddProductToProductSetOperator'
        ),
        'CloudVisionCreateProductOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionCreateProductOperator'
        ),
        'CloudVisionCreateProductSetOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionCreateProductSetOperator'
        ),
        'CloudVisionCreateReferenceImageOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionCreateReferenceImageOperator'
        ),
        'CloudVisionDeleteProductOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionDeleteProductOperator'
        ),
        'CloudVisionDeleteProductSetOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionDeleteProductSetOperator'
        ),
        'CloudVisionDetectImageLabelsOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionDetectImageLabelsOperator'
        ),
        'CloudVisionDetectImageSafeSearchOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionDetectImageSafeSearchOperator'
        ),
        'CloudVisionDetectTextOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionDetectTextOperator'
        ),
        'CloudVisionGetProductOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionGetProductOperator'
        ),
        'CloudVisionGetProductSetOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionGetProductSetOperator'
        ),
        'CloudVisionImageAnnotateOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionImageAnnotateOperator'
        ),
        'CloudVisionRemoveProductFromProductSetOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionRemoveProductFromProductSetOperator'
        ),
        'CloudVisionTextDetectOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionTextDetectOperator'
        ),
        'CloudVisionUpdateProductOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionUpdateProductOperator'
        ),
        'CloudVisionUpdateProductSetOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionUpdateProductSetOperator'
        ),
        'CloudVisionAnnotateImageOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionImageAnnotateOperator'
        ),
        'CloudVisionDetectDocumentTextOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionTextDetectOperator'
        ),
        'CloudVisionProductCreateOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionCreateProductOperator'
        ),
        'CloudVisionProductDeleteOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionDeleteProductOperator'
        ),
        'CloudVisionProductGetOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionGetProductOperator'
        ),
        'CloudVisionProductSetCreateOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionCreateProductSetOperator'
        ),
        'CloudVisionProductSetDeleteOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionDeleteProductSetOperator'
        ),
        'CloudVisionProductSetGetOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionGetProductSetOperator'
        ),
        'CloudVisionProductSetUpdateOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionUpdateProductSetOperator'
        ),
        'CloudVisionProductUpdateOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionUpdateProductOperator'
        ),
        'CloudVisionReferenceImageCreateOperator': (
            'airflow.providers.google.cloud.operators.vision.CloudVisionCreateReferenceImageOperator'
        ),
    },
    'gcs_acl_operator': {
        'GCSBucketCreateAclEntryOperator': (
            'airflow.providers.google.cloud.operators.gcs.GCSBucketCreateAclEntryOperator'
        ),
        'GCSObjectCreateAclEntryOperator': (
            'airflow.providers.google.cloud.operators.gcs.GCSObjectCreateAclEntryOperator'
        ),
        'GoogleCloudStorageBucketCreateAclEntryOperator': (
            'airflow.providers.google.cloud.operators.gcs.GCSBucketCreateAclEntryOperator'
        ),
        'GoogleCloudStorageObjectCreateAclEntryOperator': (
            'airflow.providers.google.cloud.operators.gcs.GCSObjectCreateAclEntryOperator'
        ),
    },
    'gcs_delete_operator': {
        'GCSDeleteObjectsOperator': 'airflow.providers.google.cloud.operators.gcs.GCSDeleteObjectsOperator',
        'GoogleCloudStorageDeleteOperator': (
            'airflow.providers.google.cloud.operators.gcs.GCSDeleteObjectsOperator'
        ),
    },
    'gcs_download_operator': {
        'GCSToLocalFilesystemOperator': (
            'airflow.providers.google.cloud.transfers.gcs_to_local.GCSToLocalFilesystemOperator'
        ),
        'GoogleCloudStorageDownloadOperator': (
            'airflow.providers.google.cloud.transfers.gcs_to_local.GCSToLocalFilesystemOperator'
        ),
    },
    'gcs_list_operator': {
        'GCSListObjectsOperator': 'airflow.providers.google.cloud.operators.gcs.GCSListObjectsOperator',
        'GoogleCloudStorageListOperator': (
            'airflow.providers.google.cloud.operators.gcs.GCSListObjectsOperator'
        ),
    },
    'gcs_operator': {
        'GCSCreateBucketOperator': 'airflow.providers.google.cloud.operators.gcs.GCSCreateBucketOperator',
        'GoogleCloudStorageCreateBucketOperator': (
            'airflow.providers.google.cloud.operators.gcs.GCSCreateBucketOperator'
        ),
    },
    'gcs_to_bq': {
        'GCSToBigQueryOperator': (
            'airflow.providers.google.cloud.transfers.gcs_to_bigquery.GCSToBigQueryOperator'
        ),
        'GoogleCloudStorageToBigQueryOperator': (
            'airflow.providers.google.cloud.transfers.gcs_to_bigquery.GCSToBigQueryOperator'
        ),
    },
    'gcs_to_gcs': {
        'GCSToGCSOperator': 'airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSToGCSOperator',
        'GoogleCloudStorageToGoogleCloudStorageOperator': (
            'airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSToGCSOperator'
        ),
    },
    'gcs_to_gdrive_operator': {
        'GCSToGoogleDriveOperator': (
            'airflow.providers.google.suite.transfers.gcs_to_gdrive.GCSToGoogleDriveOperator'
        ),
    },
    'gcs_to_s3': {
        'GCSToS3Operator': 'airflow.providers.amazon.aws.transfers.gcs_to_s3.GCSToS3Operator',
        'GoogleCloudStorageToS3Operator': 'airflow.providers.amazon.aws.transfers.gcs_to_s3.GCSToS3Operator',
    },
    'grpc_operator': {
        'GrpcOperator': 'airflow.providers.grpc.operators.grpc.GrpcOperator',
    },
    'hive_to_dynamodb': {
        'HiveToDynamoDBOperator': (
            'airflow.providers.amazon.aws.transfers.hive_to_dynamodb.HiveToDynamoDBOperator'
        ),
    },
    'imap_attachment_to_s3_operator': {
        'ImapAttachmentToS3Operator': (
            'airflow.providers.amazon.aws.transfers.imap_attachment_to_s3.ImapAttachmentToS3Operator'
        ),
    },
    'jenkins_job_trigger_operator': {
        'JenkinsJobTriggerOperator': (
            'airflow.providers.jenkins.operators.jenkins_job_trigger.JenkinsJobTriggerOperator'
        ),
    },
    'jira_operator': {
        'JiraOperator': 'airflow.providers.atlassian.jira.operators.jira.JiraOperator',
    },
    'kubernetes_pod_operator': {
        'KubernetesPodOperator': (
            'airflow.providers.cncf.kubernetes.operators.pod.KubernetesPodOperator'
        ),
    },
    'mlengine_operator': {
        'MLEngineManageModelOperator': (
            'airflow.providers.google.cloud.operators.mlengine.MLEngineManageModelOperator'
        ),
        'MLEngineManageVersionOperator': (
            'airflow.providers.google.cloud.operators.mlengine.MLEngineManageVersionOperator'
        ),
        'MLEngineStartBatchPredictionJobOperator': (
            'airflow.providers.google.cloud.operators.mlengine.MLEngineStartBatchPredictionJobOperator'
        ),
        'MLEngineStartTrainingJobOperator': (
            'airflow.providers.google.cloud.operators.mlengine.MLEngineStartTrainingJobOperator'
        ),
        'MLEngineBatchPredictionOperator': (
            'airflow.providers.google.cloud.operators.mlengine.MLEngineStartBatchPredictionJobOperator'
        ),
        'MLEngineModelOperator': (
            'airflow.providers.google.cloud.operators.mlengine.MLEngineManageModelOperator'
        ),
        'MLEngineTrainingOperator': (
            'airflow.providers.google.cloud.operators.mlengine.MLEngineStartTrainingJobOperator'
        ),
        'MLEngineVersionOperator': (
            'airflow.providers.google.cloud.operators.mlengine.MLEngineManageVersionOperator'
        ),
    },
    'mongo_to_s3': {
        'MongoToS3Operator': 'airflow.providers.amazon.aws.transfers.mongo_to_s3.MongoToS3Operator',
    },
    'mssql_to_gcs': {
        'MSSQLToGCSOperator': 'airflow.providers.google.cloud.transfers.mssql_to_gcs.MSSQLToGCSOperator',
        'MsSqlToGoogleCloudStorageOperator': (
            'airflow.providers.google.cloud.transfers.mssql_to_gcs.MSSQLToGCSOperator'
        ),
    },
    'mysql_to_gcs': {
        'MySQLToGCSOperator': 'airflow.providers.google.cloud.transfers.mysql_to_gcs.MySQLToGCSOperator',
        'MySqlToGoogleCloudStorageOperator': (
            'airflow.providers.google.cloud.transfers.mysql_to_gcs.MySQLToGCSOperator'
        ),
    },
    'opsgenie_alert_operator': {
        'OpsgenieCreateAlertOperator': (
            'airflow.providers.opsgenie.operators.opsgenie.OpsgenieCreateAlertOperator'
        ),
        'OpsgenieAlertOperator': 'airflow.providers.opsgenie.operators.opsgenie.OpsgenieCreateAlertOperator',
    },
    'oracle_to_azure_data_lake_transfer': {
        'OracleToAzureDataLakeOperator':
            'airflow.providers.microsoft.azure.transfers.'
            'oracle_to_azure_data_lake.OracleToAzureDataLakeOperator',
    },
    'oracle_to_oracle_transfer': {
        'OracleToOracleOperator': (
            'airflow.providers.oracle.transfers.oracle_to_oracle.OracleToOracleOperator'
        ),
        'OracleToOracleTransfer': (
            'airflow.providers.oracle.transfers.oracle_to_oracle.OracleToOracleOperator'
        ),
    },
    'postgres_to_gcs_operator': {
        'PostgresToGCSOperator': (
            'airflow.providers.google.cloud.transfers.postgres_to_gcs.PostgresToGCSOperator'
        ),
        'PostgresToGoogleCloudStorageOperator': (
            'airflow.providers.google.cloud.transfers.postgres_to_gcs.PostgresToGCSOperator'
        ),
    },
    'pubsub_operator': {
        'PubSubCreateSubscriptionOperator': (
            'airflow.providers.google.cloud.operators.pubsub.PubSubCreateSubscriptionOperator'
        ),
        'PubSubCreateTopicOperator': (
            'airflow.providers.google.cloud.operators.pubsub.PubSubCreateTopicOperator'
        ),
        'PubSubDeleteSubscriptionOperator': (
            'airflow.providers.google.cloud.operators.pubsub.PubSubDeleteSubscriptionOperator'
        ),
        'PubSubDeleteTopicOperator': (
            'airflow.providers.google.cloud.operators.pubsub.PubSubDeleteTopicOperator'
        ),
        'PubSubPublishMessageOperator': (
            'airflow.providers.google.cloud.operators.pubsub.PubSubPublishMessageOperator'
        ),
        'PubSubPublishOperator': (
            'airflow.providers.google.cloud.operators.pubsub.PubSubPublishMessageOperator'
        ),
        'PubSubSubscriptionCreateOperator': (
            'airflow.providers.google.cloud.operators.pubsub.PubSubCreateSubscriptionOperator'
        ),
        'PubSubSubscriptionDeleteOperator': (
            'airflow.providers.google.cloud.operators.pubsub.PubSubDeleteSubscriptionOperator'
        ),
        'PubSubTopicCreateOperator': (
            'airflow.providers.google.cloud.operators.pubsub.PubSubCreateTopicOperator'
        ),
        'PubSubTopicDeleteOperator': (
            'airflow.providers.google.cloud.operators.pubsub.PubSubDeleteTopicOperator'
        ),
    },
    'qubole_check_operator': {
        'QuboleCheckOperator': 'airflow.providers.qubole.operators.qubole_check.QuboleCheckOperator',
        'QuboleValueCheckOperator': (
            'airflow.providers.qubole.operators.qubole_check.QuboleValueCheckOperator'
        ),
    },
    'qubole_operator': {
        'QuboleOperator': 'airflow.providers.qubole.operators.qubole.QuboleOperator',
    },
    'redis_publish_operator': {
        'RedisPublishOperator': 'airflow.providers.redis.operators.redis_publish.RedisPublishOperator',
    },
    's3_to_gcs_operator': {
        'S3ToGCSOperator': 'airflow.providers.google.cloud.transfers.s3_to_gcs.S3ToGCSOperator',
    },
    's3_to_gcs_transfer_operator': {
        'CloudDataTransferServiceS3ToGCSOperator':
            'airflow.providers.google.cloud.operators.cloud_storage_transfer_service.'
            'CloudDataTransferServiceS3ToGCSOperator',
    },
    's3_to_sftp_operator': {
        'S3ToSFTPOperator': 'airflow.providers.amazon.aws.transfers.s3_to_sftp.S3ToSFTPOperator',
    },
    'segment_track_event_operator': {
        'SegmentTrackEventOperator': (
            'airflow.providers.segment.operators.segment_track_event.SegmentTrackEventOperator'
        ),
    },
    'sftp_operator': {
        'SFTPOperator': 'airflow.providers.sftp.operators.sftp.SFTPOperator',
    },
    'sftp_to_s3_operator': {
        'SFTPToS3Operator': 'airflow.providers.amazon.aws.transfers.sftp_to_s3.SFTPToS3Operator',
    },
    'slack_webhook_operator': {
        'SlackWebhookOperator': 'airflow.providers.slack.operators.slack_webhook.SlackWebhookOperator',
    },
    'snowflake_operator': {
        'SnowflakeOperator': 'airflow.providers.snowflake.operators.snowflake.SnowflakeOperator',
    },
    'sns_publish_operator': {
        'SnsPublishOperator': 'airflow.providers.amazon.aws.operators.sns.SnsPublishOperator',
    },
    'spark_jdbc_operator': {
        'SparkJDBCOperator': 'airflow.providers.apache.spark.operators.spark_jdbc.SparkJDBCOperator',
        'SparkSubmitOperator': 'airflow.providers.apache.spark.operators.spark_jdbc.SparkSubmitOperator',
    },
    'spark_sql_operator': {
        'SparkSqlOperator': 'airflow.providers.apache.spark.operators.spark_sql.SparkSqlOperator',
    },
    'spark_submit_operator': {
        'SparkSubmitOperator': 'airflow.providers.apache.spark.operators.spark_submit.SparkSubmitOperator',
    },
    'sql_to_gcs': {
        'BaseSQLToGCSOperator': 'airflow.providers.google.cloud.transfers.sql_to_gcs.BaseSQLToGCSOperator',
        'BaseSQLToGoogleCloudStorageOperator': (
            'airflow.providers.google.cloud.transfers.sql_to_gcs.BaseSQLToGCSOperator'
        ),
    },
    'sqoop_operator': {
        'SqoopOperator': 'airflow.providers.apache.sqoop.operators.sqoop.SqoopOperator',
    },
    'ssh_operator': {
        'SSHOperator': 'airflow.providers.ssh.operators.ssh.SSHOperator',
    },
    'vertica_operator': {
        'VerticaOperator': 'airflow.providers.vertica.operators.vertica.VerticaOperator',
    },
    'vertica_to_hive': {
        'VerticaToHiveOperator': (
            'airflow.providers.apache.hive.transfers.vertica_to_hive.VerticaToHiveOperator'
        ),
        'VerticaToHiveTransfer': (
            'airflow.providers.apache.hive.transfers.vertica_to_hive.VerticaToHiveOperator'
        ),
    },
    'vertica_to_mysql': {
        'VerticaToMySqlOperator': 'airflow.providers.mysql.transfers.vertica_to_mysql.VerticaToMySqlOperator',
        'VerticaToMySqlTransfer': 'airflow.providers.mysql.transfers.vertica_to_mysql.VerticaToMySqlOperator',
    },
    'wasb_delete_blob_operator': {
        'WasbDeleteBlobOperator': (
            'airflow.providers.microsoft.azure.operators.wasb_delete_blob.WasbDeleteBlobOperator'
        ),
    },
    'winrm_operator': {
        'WinRMOperator': 'airflow.providers.microsoft.winrm.operators.winrm.WinRMOperator',
    },
}

add_deprecated_classes(__deprecated_classes, __name__)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Task sub-commands."""
from __future__ import annotations

import importlib
import json
import logging
import os
import sys
import textwrap
from contextlib import contextmanager, redirect_stderr, redirect_stdout, suppress
from typing import Generator, Union, cast

import pendulum
from pendulum.parsing.exceptions import ParserError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.session import Session

from airflow import settings
from airflow.cli.simple_table import AirflowConsole
from airflow.configuration import conf
from airflow.exceptions import AirflowException, DagRunNotFound, TaskInstanceNotFound
from airflow.executors.executor_loader import ExecutorLoader
from airflow.jobs.job import Job, run_job
from airflow.jobs.local_task_job_runner import LocalTaskJobRunner
from airflow.listeners.listener import get_listener_manager
from airflow.models import DagPickle, TaskInstance
from airflow.models.dag import DAG
from airflow.models.dagrun import DagRun
from airflow.models.operator import Operator, needs_expansion
from airflow.models.param import ParamsDict
from airflow.models.taskinstance import TaskReturnCode
from airflow.settings import IS_K8S_EXECUTOR_POD
from airflow.ti_deps.dep_context import DepContext
from airflow.ti_deps.dependencies_deps import SCHEDULER_QUEUED_DEPS
from airflow.typing_compat import Literal, Protocol
from airflow.utils import cli as cli_utils
from airflow.utils.cli import (
    get_dag,
    get_dag_by_file_location,
    get_dag_by_pickle,
    get_dags,
    should_ignore_depends_on_past,
    suppress_logs_and_warning,
)
from airflow.utils.dates import timezone
from airflow.utils.log.file_task_handler import _set_task_deferred_context_var
from airflow.utils.log.logging_mixin import StreamLogWriter
from airflow.utils.log.secrets_masker import RedactedIO
from airflow.utils.net import get_hostname
from airflow.utils.session import NEW_SESSION, create_session, provide_session
from airflow.utils.state import DagRunState

log = logging.getLogger(__name__)

CreateIfNecessary = Union[Literal[False], Literal["db"], Literal["memory"]]


def _generate_temporary_run_id() -> str:
    """Generate a ``run_id`` for a DAG run that will be created temporarily.

    This is used mostly by ``airflow task test`` to create a DAG run that will
    be deleted after the task is run.
    """
    return f"__airflow_temporary_run_{timezone.utcnow().isoformat()}__"


def _get_dag_run(
    *,
    dag: DAG,
    create_if_necessary: CreateIfNecessary,
    exec_date_or_run_id: str | None = None,
    session: Session,
) -> tuple[DagRun, bool]:
    """Try to retrieve a DAG run from a string representing either a run ID or logical date.

    This checks DAG runs like this:

    1. If the input ``exec_date_or_run_id`` matches a DAG run ID, return the run.
    2. Try to parse the input as a date. If that works, and the resulting
       date matches a DAG run's logical date, return the run.
    3. If ``create_if_necessary`` is *False* and the input works for neither of
       the above, raise ``DagRunNotFound``.
    4. Try to create a new DAG run. If the input looks like a date, use it as
       the logical date; otherwise use it as a run ID and set the logical date
       to the current time.
    """
    if not exec_date_or_run_id and not create_if_necessary:
        raise ValueError("Must provide `exec_date_or_run_id` if not `create_if_necessary`.")
    execution_date: pendulum.DateTime | None = None
    if exec_date_or_run_id:
        dag_run = dag.get_dagrun(run_id=exec_date_or_run_id, session=session)
        if dag_run:
            return dag_run, False
        with suppress(ParserError, TypeError):
            execution_date = timezone.parse(exec_date_or_run_id)
        try:
            dag_run = (
                session.query(DagRun)
                .filter(DagRun.dag_id == dag.dag_id, DagRun.execution_date == execution_date)
                .one()
            )
        except NoResultFound:
            if not create_if_necessary:
                raise DagRunNotFound(
                    f"DagRun for {dag.dag_id} with run_id or execution_date "
                    f"of {exec_date_or_run_id!r} not found"
                ) from None
        else:
            return dag_run, False

    if execution_date is not None:
        dag_run_execution_date = execution_date
    else:
        dag_run_execution_date = pendulum.instance(timezone.utcnow())

    if create_if_necessary == "memory":
        dag_run = DagRun(dag.dag_id, run_id=exec_date_or_run_id, execution_date=dag_run_execution_date)
        return dag_run, True
    elif create_if_necessary == "db":
        dag_run = dag.create_dagrun(
            state=DagRunState.QUEUED,
            execution_date=dag_run_execution_date,
            run_id=_generate_temporary_run_id(),
            data_interval=dag.timetable.infer_manual_data_interval(run_after=dag_run_execution_date),
            session=session,
        )
        return dag_run, True
    raise ValueError(f"unknown create_if_necessary value: {create_if_necessary!r}")


@provide_session
def _get_ti(
    task: Operator,
    map_index: int,
    *,
    exec_date_or_run_id: str | None = None,
    pool: str | None = None,
    create_if_necessary: CreateIfNecessary = False,
    session: Session = NEW_SESSION,
) -> tuple[TaskInstance, bool]:
    """Get the task instance through DagRun.run_id, if that fails, get the TI the old way."""
    dag = task.dag
    if dag is None:
        raise ValueError("Cannot get task instance for a task not assigned to a DAG")
    if not exec_date_or_run_id and not create_if_necessary:
        raise ValueError("Must provide `exec_date_or_run_id` if not `create_if_necessary`.")
    if needs_expansion(task):
        if map_index < 0:
            raise RuntimeError("No map_index passed to mapped task")
    elif map_index >= 0:
        raise RuntimeError("map_index passed to non-mapped task")
    dag_run, dr_created = _get_dag_run(
        dag=dag,
        exec_date_or_run_id=exec_date_or_run_id,
        create_if_necessary=create_if_necessary,
        session=session,
    )

    ti_or_none = dag_run.get_task_instance(task.task_id, map_index=map_index, session=session)
    if ti_or_none is None:
        if not create_if_necessary:
            raise TaskInstanceNotFound(
                f"TaskInstance for {dag.dag_id}, {task.task_id}, map={map_index} with "
                f"run_id or execution_date of {exec_date_or_run_id!r} not found"
            )
        # TODO: Validate map_index is in range?
        ti = TaskInstance(task, run_id=dag_run.run_id, map_index=map_index)
        ti.dag_run = dag_run
    else:
        ti = ti_or_none
    ti.refresh_from_task(task, pool_override=pool)
    return ti, dr_created


def _run_task_by_selected_method(args, dag: DAG, ti: TaskInstance) -> None | TaskReturnCode:
    """
    Runs the task based on a mode.

    Any of the 3 modes are available:

    - using LocalTaskJob
    - as raw task
    - by executor
    """
    if args.local:
        return _run_task_by_local_task_job(args, ti)
    if args.raw:
        return _run_raw_task(args, ti)
    _run_task_by_executor(args, dag, ti)
    return None


def _run_task_by_executor(args, dag: DAG, ti: TaskInstance) -> None:
    """
    Sends the task to the executor for execution.

    This can result in the task being started by another host if the executor implementation does.
    """
    pickle_id = None
    if args.ship_dag:
        try:
            # Running remotely, so pickling the DAG
            with create_session() as session:
                pickle = DagPickle(dag)
                session.add(pickle)
            pickle_id = pickle.id
            # TODO: This should be written to a log
            print(f"Pickled dag {dag} as pickle_id: {pickle_id}")
        except Exception as e:
            print("Could not pickle the DAG")
            print(e)
            raise e
    executor = ExecutorLoader.get_default_executor()
    executor.job_id = None
    executor.start()
    print("Sending to executor.")
    executor.queue_task_instance(
        ti,
        mark_success=args.mark_success,
        pickle_id=pickle_id,
        ignore_all_deps=args.ignore_all_dependencies,
        ignore_depends_on_past=should_ignore_depends_on_past(args),
        wait_for_past_depends_before_skipping=(args.depends_on_past == "wait"),
        ignore_task_deps=args.ignore_dependencies,
        ignore_ti_state=args.force,
        pool=args.pool,
    )
    executor.heartbeat()
    executor.end()


def _run_task_by_local_task_job(args, ti: TaskInstance) -> TaskReturnCode | None:
    """Run LocalTaskJob, which monitors the raw task execution process."""
    job_runner = LocalTaskJobRunner(
        job=Job(dag_id=ti.dag_id),
        task_instance=ti,
        mark_success=args.mark_success,
        pickle_id=args.pickle,
        ignore_all_deps=args.ignore_all_dependencies,
        ignore_depends_on_past=should_ignore_depends_on_past(args),
        wait_for_past_depends_before_skipping=(args.depends_on_past == "wait"),
        ignore_task_deps=args.ignore_dependencies,
        ignore_ti_state=args.force,
        pool=args.pool,
        external_executor_id=_extract_external_executor_id(args),
    )
    try:
        ret = run_job(job=job_runner.job, execute_callable=job_runner._execute)
    finally:
        if args.shut_down_logging:
            logging.shutdown()
    with suppress(ValueError):
        return TaskReturnCode(ret)
    return None


RAW_TASK_UNSUPPORTED_OPTION = [
    "ignore_all_dependencies",
    "ignore_depends_on_past",
    "ignore_dependencies",
    "force",
]


def _run_raw_task(args, ti: TaskInstance) -> None | TaskReturnCode:
    """Runs the main task handling code."""
    return ti._run_raw_task(
        mark_success=args.mark_success,
        job_id=args.job_id,
        pool=args.pool,
    )


def _extract_external_executor_id(args) -> str | None:
    if hasattr(args, "external_executor_id"):
        return getattr(args, "external_executor_id")
    return os.environ.get("external_executor_id", None)


@contextmanager
def _move_task_handlers_to_root(ti: TaskInstance) -> Generator[None, None, None]:
    """
    Move handlers for task logging to root logger.

    We want anything logged during task run to be propagated to task log handlers.
    If running in a k8s executor pod, also keep the stream handler on root logger
    so that logs are still emitted to stdout.
    """
    # nothing to do
    if not ti.log.handlers or settings.DONOT_MODIFY_HANDLERS:
        yield
        return

    # Move task handlers to root and reset task logger and restore original logger settings after exit.
    # If k8s executor, we need to ensure that root logger has a console handler, so that
    # task logs propagate to stdout (this is how webserver retrieves them while task is running).
    root_logger = logging.getLogger()
    console_handler = next((h for h in root_logger.handlers if h.name == "console"), None)
    with LoggerMutationHelper(root_logger), LoggerMutationHelper(ti.log) as task_helper:
        task_helper.move(root_logger)
        if IS_K8S_EXECUTOR_POD:
            if console_handler and console_handler not in root_logger.handlers:
                root_logger.addHandler(console_handler)
        yield


@contextmanager
def _redirect_stdout_to_ti_log(ti: TaskInstance) -> Generator[None, None, None]:
    """
    Redirect stdout to ti logger.

    Redirect stdout and stderr to the task instance log as INFO and WARNING
    level messages, respectively.

    If stdout already redirected (possible when task running with option
    `--local`), don't redirect again.
    """
    # if sys.stdout is StreamLogWriter, it means we already redirected
    # likely before forking in LocalTaskJob
    if not isinstance(sys.stdout, StreamLogWriter):
        info_writer = StreamLogWriter(ti.log, logging.INFO)
        warning_writer = StreamLogWriter(ti.log, logging.WARNING)
        with redirect_stdout(info_writer), redirect_stderr(warning_writer):
            yield
    else:
        yield


class TaskCommandMarker:
    """Marker for listener hooks, to properly detect from which component they are called."""


@cli_utils.action_cli(check_db=False)
def task_run(args, dag: DAG | None = None) -> TaskReturnCode | None:
    """
    Run a single task instance.

    Note that there must be at least one DagRun for this to start,
    i.e. it must have been scheduled and/or triggered previously.
    Alternatively, if you just need to run it for testing then use
    "airflow tasks test ..." command instead.
    """
    # Load custom airflow config

    if args.local and args.raw:
        raise AirflowException(
            "Option --raw and --local are mutually exclusive. "
            "Please remove one option to execute the command."
        )

    if args.raw:
        unsupported_options = [o for o in RAW_TASK_UNSUPPORTED_OPTION if getattr(args, o)]

        if unsupported_options:
            unsupported_raw_task_flags = ", ".join(f"--{o}" for o in RAW_TASK_UNSUPPORTED_OPTION)
            unsupported_flags = ", ".join(f"--{o}" for o in unsupported_options)
            raise AirflowException(
                "Option --raw does not work with some of the other options on this command. "
                "You can't use --raw option and the following options: "
                f"{unsupported_raw_task_flags}. "
                f"You provided the option {unsupported_flags}. "
                "Delete it to execute the command."
            )
    if dag and args.pickle:
        raise AirflowException("You cannot use the --pickle option when using DAG.cli() method.")
    if args.cfg_path:
        with open(args.cfg_path) as conf_file:
            conf_dict = json.load(conf_file)

        if os.path.exists(args.cfg_path):
            os.remove(args.cfg_path)

        conf.read_dict(conf_dict, source=args.cfg_path)
        settings.configure_vars()

    settings.MASK_SECRETS_IN_LOGS = True

    get_listener_manager().hook.on_starting(component=TaskCommandMarker())

    if args.pickle:
        print(f"Loading pickle id: {args.pickle}")
        _dag = get_dag_by_pickle(args.pickle)
    elif not dag:
        _dag = get_dag(args.subdir, args.dag_id)
    else:
        _dag = dag
    task = _dag.get_task(task_id=args.task_id)
    ti, _ = _get_ti(task, args.map_index, exec_date_or_run_id=args.execution_date_or_run_id, pool=args.pool)
    ti.init_run_context(raw=args.raw)

    hostname = get_hostname()

    log.info("Running %s on host %s", ti, hostname)

    # IMPORTANT, have to re-configure ORM with the NullPool, otherwise, each "run" command may leave
    # behind multiple open sleeping connections while heartbeating, which could
    # easily exceed the database connection limit when
    # processing hundreds of simultaneous tasks.
    # this should be last thing before running, to reduce likelihood of an open session
    # which can cause trouble if running process in a fork.
    settings.reconfigure_orm(disable_connection_pool=True)
    task_return_code = None
    try:
        if args.interactive:
            task_return_code = _run_task_by_selected_method(args, _dag, ti)
        else:
            with _move_task_handlers_to_root(ti), _redirect_stdout_to_ti_log(ti):
                task_return_code = _run_task_by_selected_method(args, _dag, ti)
                if task_return_code == TaskReturnCode.DEFERRED:
                    _set_task_deferred_context_var()
    finally:
        try:
            get_listener_manager().hook.before_stopping(component=TaskCommandMarker())
        except Exception:
            pass
    return task_return_code


@cli_utils.action_cli(check_db=False)
def task_failed_deps(args) -> None:
    """
    Get task instance dependencies that were not met.

    Returns the unmet dependencies for a task instance from the perspective of the
    scheduler (i.e. why a task instance doesn't get scheduled and then queued by the
    scheduler, and then run by an executor).
    >>> airflow tasks failed-deps tutorial sleep 2015-01-01
    Task instance dependencies not met:
    Dagrun Running: Task instance's dagrun did not exist: Unknown reason
    Trigger Rule: Task's trigger rule 'all_success' requires all upstream tasks
    to have succeeded, but found 1 non-success(es).
    """
    dag = get_dag(args.subdir, args.dag_id)
    task = dag.get_task(task_id=args.task_id)
    ti, _ = _get_ti(task, args.map_index, exec_date_or_run_id=args.execution_date_or_run_id)

    dep_context = DepContext(deps=SCHEDULER_QUEUED_DEPS)
    failed_deps = list(ti.get_failed_dep_statuses(dep_context=dep_context))
    # TODO, Do we want to print or log this
    if failed_deps:
        print("Task instance dependencies not met:")
        for dep in failed_deps:
            print(f"{dep.dep_name}: {dep.reason}")
    else:
        print("Task instance dependencies are all met.")


@cli_utils.action_cli(check_db=False)
@suppress_logs_and_warning
def task_state(args) -> None:
    """
    Returns the state of a TaskInstance at the command line.
    >>> airflow tasks state tutorial sleep 2015-01-01
    success
    """
    dag = get_dag(args.subdir, args.dag_id)
    task = dag.get_task(task_id=args.task_id)
    ti, _ = _get_ti(task, args.map_index, exec_date_or_run_id=args.execution_date_or_run_id)
    print(ti.current_state())


@cli_utils.action_cli(check_db=False)
@suppress_logs_and_warning
def task_list(args, dag: DAG | None = None) -> None:
    """Lists the tasks within a DAG at the command line."""
    dag = dag or get_dag(args.subdir, args.dag_id)
    if args.tree:
        dag.tree_view()
    else:
        tasks = sorted(t.task_id for t in dag.tasks)
        print("\n".join(tasks))


class _SupportedDebugger(Protocol):
    def post_mortem(self) -> None:
        ...


SUPPORTED_DEBUGGER_MODULES = [
    "pudb",
    "web_pdb",
    "ipdb",
    "pdb",
]


def _guess_debugger() -> _SupportedDebugger:
    """
    Trying to guess the debugger used by the user.

    When it doesn't find any user-installed debugger, returns ``pdb``.

    List of supported debuggers:

    * `pudb <https://github.com/inducer/pudb>`__
    * `web_pdb <https://github.com/romanvm/python-web-pdb>`__
    * `ipdb <https://github.com/gotcha/ipdb>`__
    * `pdb <https://docs.python.org/3/library/pdb.html>`__
    """
    exc: Exception
    for mod_name in SUPPORTED_DEBUGGER_MODULES:
        try:
            return cast(_SupportedDebugger, importlib.import_module(mod_name))
        except ImportError as e:
            exc = e
    raise exc


@cli_utils.action_cli(check_db=False)
@suppress_logs_and_warning
@provide_session
def task_states_for_dag_run(args, session: Session = NEW_SESSION) -> None:
    """Get the status of all task instances in a DagRun."""
    dag_run = (
        session.query(DagRun)
        .filter(DagRun.run_id == args.execution_date_or_run_id, DagRun.dag_id == args.dag_id)
        .one_or_none()
    )
    if not dag_run:
        try:
            execution_date = timezone.parse(args.execution_date_or_run_id)
            dag_run = (
                session.query(DagRun)
                .filter(DagRun.execution_date == execution_date, DagRun.dag_id == args.dag_id)
                .one_or_none()
            )
        except (ParserError, TypeError) as err:
            raise AirflowException(f"Error parsing the supplied execution_date. Error: {str(err)}")

    if dag_run is None:
        raise DagRunNotFound(
            f"DagRun for {args.dag_id} with run_id or execution_date of {args.execution_date_or_run_id!r} "
            "not found"
        )

    has_mapped_instances = any(ti.map_index >= 0 for ti in dag_run.task_instances)

    def format_task_instance(ti: TaskInstance) -> dict[str, str]:
        data = {
            "dag_id": ti.dag_id,
            "execution_date": dag_run.execution_date.isoformat(),
            "task_id": ti.task_id,
            "state": ti.state,
            "start_date": ti.start_date.isoformat() if ti.start_date else "",
            "end_date": ti.end_date.isoformat() if ti.end_date else "",
        }
        if has_mapped_instances:
            data["map_index"] = str(ti.map_index) if ti.map_index >= 0 else ""
        return data

    AirflowConsole().print_as(data=dag_run.task_instances, output=args.output, mapper=format_task_instance)


@cli_utils.action_cli(check_db=False)
def task_test(args, dag: DAG | None = None) -> None:
    """Tests task for a given dag_id."""
    # We want to log output from operators etc to show up here. Normally
    # airflow.task would redirect to a file, but here we want it to propagate
    # up to the normal airflow handler.

    settings.MASK_SECRETS_IN_LOGS = True

    handlers = logging.getLogger("airflow.task").handlers
    already_has_stream_handler = False
    for handler in handlers:
        already_has_stream_handler = isinstance(handler, logging.StreamHandler)
        if already_has_stream_handler:
            break
    if not already_has_stream_handler:
        logging.getLogger("airflow.task").propagate = True

    env_vars = {"AIRFLOW_TEST_MODE": "True"}
    if args.env_vars:
        env_vars.update(args.env_vars)
        os.environ.update(env_vars)

    dag = dag or get_dag(args.subdir, args.dag_id)

    task = dag.get_task(task_id=args.task_id)
    # Add CLI provided task_params to task.params
    if args.task_params:
        passed_in_params = json.loads(args.task_params)
        task.params.update(passed_in_params)

    if task.params and isinstance(task.params, ParamsDict):
        task.params.validate()

    ti, dr_created = _get_ti(
        task, args.map_index, exec_date_or_run_id=args.execution_date_or_run_id, create_if_necessary="db"
    )

    try:
        with redirect_stdout(RedactedIO()):
            if args.dry_run:
                ti.dry_run()
            else:
                ti.run(ignore_task_deps=True, ignore_ti_state=True, test_mode=True)
    except Exception:
        if args.post_mortem:
            debugger = _guess_debugger()
            debugger.post_mortem()
        else:
            raise
    finally:
        if not already_has_stream_handler:
            # Make sure to reset back to normal. When run for CLI this doesn't
            # matter, but it does for test suite
            logging.getLogger("airflow.task").propagate = False
        if dr_created:
            with create_session() as session:
                session.delete(ti.dag_run)


@cli_utils.action_cli(check_db=False)
@suppress_logs_and_warning
def task_render(args, dag: DAG | None = None) -> None:
    """Renders and displays templated fields for a given task."""
    if not dag:
        dag = get_dag(args.subdir, args.dag_id)
    task = dag.get_task(task_id=args.task_id)
    ti, _ = _get_ti(
        task, args.map_index, exec_date_or_run_id=args.execution_date_or_run_id, create_if_necessary="memory"
    )
    ti.render_templates()
    for attr in task.template_fields:
        print(
            textwrap.dedent(
                f"""        # ----------------------------------------------------------
        # property: {attr}
        # ----------------------------------------------------------
        {getattr(ti.task, attr)}
        """
            )
        )


@cli_utils.action_cli(check_db=False)
def task_clear(args) -> None:
    """Clears all task instances or only those matched by regex for a DAG(s)."""
    logging.basicConfig(level=settings.LOGGING_LEVEL, format=settings.SIMPLE_LOG_FORMAT)

    if args.dag_id and not args.subdir and not args.dag_regex and not args.task_regex:
        dags = [get_dag_by_file_location(args.dag_id)]
    else:
        # todo clear command only accepts a single dag_id. no reason for get_dags with 's' except regex?
        dags = get_dags(args.subdir, args.dag_id, use_regex=args.dag_regex)

        if args.task_regex:
            for idx, dag in enumerate(dags):
                dags[idx] = dag.partial_subset(
                    task_ids_or_regex=args.task_regex,
                    include_downstream=args.downstream,
                    include_upstream=args.upstream,
                )

    DAG.clear_dags(
        dags,
        start_date=args.start_date,
        end_date=args.end_date,
        only_failed=args.only_failed,
        only_running=args.only_running,
        confirm_prompt=not args.yes,
        include_subdags=not args.exclude_subdags,
        include_parentdag=not args.exclude_parentdag,
    )


class LoggerMutationHelper:
    """
    Helper for moving and resetting handlers and other logger attrs.

    :meta private:
    """

    def __init__(self, logger: logging.Logger) -> None:
        self.handlers = logger.handlers[:]
        self.level = logger.level
        self.propagate = logger.propagate
        self.source_logger = logger

    def apply(self, logger: logging.Logger, replace: bool = True) -> None:
        """
        Set ``logger`` with attrs stored on instance.

        If ``logger`` is root logger, don't change propagate.
        """
        if replace:
            logger.handlers[:] = self.handlers
        else:
            for h in self.handlers:
                if h not in logger.handlers:
                    logger.addHandler(h)
        logger.level = self.level
        if logger is not logging.getLogger():
            logger.propagate = self.propagate

    def move(self, logger: logging.Logger, replace: bool = True) -> None:
        """
        Replace ``logger`` attrs with those from source.

        :param logger: target logger
        :param replace: if True, remove all handlers from target first; otherwise add if not present.
        """
        self.apply(logger, replace=replace)
        self.source_logger.propagate = True
        self.source_logger.handlers[:] = []

    def reset(self) -> None:
        self.apply(self.source_logger)

    def __enter__(self) -> LoggerMutationHelper:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.reset()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
from subprocess import PIPE, STDOUT, Popen
from tempfile import NamedTemporaryFile, TemporaryDirectory, gettempdir
from typing import Sequence

from airflow.exceptions import AirflowFailException
from airflow.sensors.base import BaseSensorOperator
from airflow.utils.context import Context


class BashSensor(BaseSensorOperator):
    """
    Executes a bash command/script.

    Return True if and only if the return code is 0.

    :param bash_command: The command, set of commands or reference to a
        bash script (must be '.sh') to be executed.

    :param env: If env is not None, it must be a mapping that defines the
        environment variables for the new process; these are used instead
        of inheriting the current process environment, which is the default
        behavior. (templated)
    :param output_encoding: output encoding of bash command.
    :param retry_exit_code: If task exits with this code, treat the sensor
        as not-yet-complete and retry the check later according to the
        usual retry/timeout settings. Any other non-zero return code will
        be treated as an error, and cause the sensor to fail. If set to
        ``None`` (the default), any non-zero exit code will cause a retry
        and the task will never raise an error except on time-out.

    .. seealso::
        For more information on how to use this sensor,take a look at the guide:
        :ref:`howto/operator:BashSensor`
    """

    template_fields: Sequence[str] = ("bash_command", "env")

    def __init__(
        self, *, bash_command, env=None, output_encoding="utf-8", retry_exit_code: int | None = None, **kwargs
    ):
        super().__init__(**kwargs)
        self.bash_command = bash_command
        self.env = env
        self.output_encoding = output_encoding
        self.retry_exit_code = retry_exit_code

    def poke(self, context: Context):
        """Execute the bash command in a temporary directory."""
        bash_command = self.bash_command
        self.log.info("Tmp dir root location: \n %s", gettempdir())
        with TemporaryDirectory(prefix="airflowtmp") as tmp_dir:
            with NamedTemporaryFile(dir=tmp_dir, prefix=self.task_id) as f:
                f.write(bytes(bash_command, "utf_8"))
                f.flush()
                fname = f.name
                script_location = tmp_dir + "/" + fname
                self.log.info("Temporary script location: %s", script_location)
                self.log.info("Running command: %s", bash_command)

                with Popen(
                    ["bash", fname],
                    stdout=PIPE,
                    stderr=STDOUT,
                    close_fds=True,
                    cwd=tmp_dir,
                    env=self.env,
                    preexec_fn=os.setsid,
                ) as resp:
                    if resp.stdout:
                        self.log.info("Output:")
                        for line in iter(resp.stdout.readline, b""):
                            self.log.info(line.decode(self.output_encoding).strip())
                    resp.wait()
                    self.log.info("Command exited with return code %s", resp.returncode)

                    # zero code means success, the sensor can go green
                    if resp.returncode == 0:
                        return True

                    # we have a retry exit code, sensor retries if return code matches, otherwise error
                    elif self.retry_exit_code is not None:
                        if resp.returncode == self.retry_exit_code:
                            self.log.info("Return code matches retry code, will retry later")
                            return False
                        else:
                            raise AirflowFailException(f"Command exited with return code {resp.returncode}")

                    # backwards compatibility: sensor retries no matter the error code
                    else:
                        self.log.info("Non-zero return code and no retry code set, will retry later")
                        return False



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import os
from tempfile import mkstemp

from airflow.configuration import conf
from airflow.utils.platform import IS_WINDOWS


def tmp_configuration_copy(chmod=0o600, include_env=True, include_cmds=True):
    """
    Returns a path for a temporary file including a full copy of the configuration
    settings.

    :param include_env: Should the value of configuration from ``AIRFLOW__``
        environment variables be included or not
    :param include_cmds: Should the result of calling any *_cmd config be
        set (True, default), or should the _cmd options be left as the
        command to run (False)
    :return: a path to a temporary file
    """
    cfg_dict = conf.as_dict(
        display_sensitive=True, raw=True, include_cmds=include_cmds, include_env=include_env
    )
    temp_fd, cfg_path = mkstemp()

    with os.fdopen(temp_fd, "w") as temp_file:
        # Set the permissions before we write anything to it.
        if chmod is not None and not IS_WINDOWS:
            os.fchmod(temp_fd, chmod)
        json.dump(cfg_dict, temp_file)

    return cfg_path



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

import attr
from sqlalchemy.orm.session import Session

from airflow.exceptions import TaskNotFound
from airflow.utils.state import State

if TYPE_CHECKING:
    from airflow.models.dagrun import DagRun
    from airflow.models.taskinstance import TaskInstance


@attr.define
class DepContext:
    """
    A base class for dependency contexts.

    Specifies which dependencies should be evaluated in the context for a task
    instance to satisfy the requirements of the context. Also stores state
    related to the context that can be used by dependency classes.

    For example there could be a SomeRunContext that subclasses this class which has
    dependencies for:

    - Making sure there are slots available on the infrastructure to run the task instance
    - A task-instance's task-specific dependencies are met (e.g. the previous task
      instance completed successfully)
    - ...

    :param deps: The context-specific dependencies that need to be evaluated for a
        task instance to run in this execution context.
    :param flag_upstream_failed: This is a hack to generate the upstream_failed state
        creation while checking to see whether the task instance is runnable. It was the
        shortest path to add the feature. This is bad since this class should be pure (no
        side effects).
    :param ignore_all_deps: Whether or not the context should ignore all ignorable
        dependencies. Overrides the other ignore_* parameters
    :param ignore_depends_on_past: Ignore depends_on_past parameter of DAGs (e.g. for
        Backfills)
    :param wait_for_past_depends_before_skipping: Wait for past depends before marking the ti as skipped
    :param ignore_in_retry_period: Ignore the retry period for task instances
    :param ignore_in_reschedule_period: Ignore the reschedule period for task instances
    :param ignore_unmapped_tasks: Ignore errors about mapped tasks not yet being expanded
    :param ignore_task_deps: Ignore task-specific dependencies such as depends_on_past and
        trigger rule
    :param ignore_ti_state: Ignore the task instance's previous failure/success
    :param finished_tis: A list of all the finished task instances of this run
    """

    deps: set = attr.ib(factory=set)
    flag_upstream_failed: bool = False
    ignore_all_deps: bool = False
    ignore_depends_on_past: bool = False
    wait_for_past_depends_before_skipping: bool = False
    ignore_in_retry_period: bool = False
    ignore_in_reschedule_period: bool = False
    ignore_task_deps: bool = False
    ignore_ti_state: bool = False
    ignore_unmapped_tasks: bool = False
    finished_tis: list[TaskInstance] | None = None
    description: str | None = None

    have_changed_ti_states: bool = False
    """Have any of the TIs state's been changed as a result of evaluating dependencies"""

    def ensure_finished_tis(self, dag_run: DagRun, session: Session) -> list[TaskInstance]:
        """
        This method makes sure finished_tis is populated if it's currently None.
        This is for the strange feature of running tasks without dag_run.

        :param dag_run: The DagRun for which to find finished tasks
        :return: A list of all the finished tasks of this DAG and execution_date
        """
        if self.finished_tis is None:
            finished_tis = dag_run.get_task_instances(state=State.finished, session=session)
            for ti in finished_tis:
                if not hasattr(ti, "task") and dag_run.dag:
                    try:
                        ti.task = dag_run.dag.get_task(ti.task_id)
                    except TaskNotFound:
                        pass

            self.finished_tis = finished_tis
        else:
            finished_tis = self.finished_tis
        return finished_tis



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Any

from sqlalchemy import MetaData, String
from sqlalchemy.orm import declarative_base

from airflow.configuration import conf

SQL_ALCHEMY_SCHEMA = conf.get("database", "SQL_ALCHEMY_SCHEMA")

# For more information about what the tokens in the naming convention
# below mean, see:
# https://docs.sqlalchemy.org/en/14/core/metadata.html#sqlalchemy.schema.MetaData.params.naming_convention
naming_convention = {
    "ix": "idx_%(column_0_N_label)s",
    "uq": "%(table_name)s_%(column_0_N_name)s_uq",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "%(table_name)s_%(column_0_name)s_fkey",
    "pk": "%(table_name)s_pkey",
}


def _get_schema():
    if not SQL_ALCHEMY_SCHEMA or SQL_ALCHEMY_SCHEMA.isspace():
        return None
    return SQL_ALCHEMY_SCHEMA


metadata = MetaData(schema=_get_schema(), naming_convention=naming_convention)

Base: Any = declarative_base(metadata=metadata)

ID_LEN = 250


def get_id_collation_args():
    """Get SQLAlchemy args to use for COLLATION"""
    collation = conf.get("database", "sql_engine_collation_for_ids", fallback=None)
    if collation:
        return {"collation": collation}
    else:
        # Automatically use utf8mb3_bin collation for mysql
        # This is backwards-compatible. All our IDS are ASCII anyway so even if
        # we migrate from previously installed database with different collation and we end up mixture of
        # COLLATIONS, it's not a problem whatsoever (and we keep it small enough so that our indexes
        # for MYSQL will not exceed the maximum index size.
        #
        # See https://github.com/apache/airflow/pull/17603#issuecomment-901121618.
        #
        # We cannot use session/dialect as at this point we are trying to determine the right connection
        # parameters, so we use the connection
        conn = conf.get("database", "sql_alchemy_conn", fallback="")
        if conn.startswith("mysql") or conn.startswith("mariadb"):
            return {"collation": "utf8mb3_bin"}
        return {}


COLLATION_ARGS = get_id_collation_args()


def StringID(*, length=ID_LEN, **kwargs) -> String:
    return String(length=length, **kwargs, **COLLATION_ARGS)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""add callback request table

Revision ID: c97c2ab6aa23
Revises: c306b5b5ae4a
Create Date: 2022-01-28 21:11:11.857010

"""
from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy import func

from airflow.migrations.db_types import TIMESTAMP
from airflow.utils.sqlalchemy import ExtendedJSON

# revision identifiers, used by Alembic.
revision = "c97c2ab6aa23"
down_revision = "c306b5b5ae4a"
branch_labels = None
depends_on = None
airflow_version = "2.3.0"

TABLE_NAME = "callback_request"


def upgrade():
    op.create_table(
        TABLE_NAME,
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("created_at", TIMESTAMP, default=func.now, nullable=False),
        sa.Column("priority_weight", sa.Integer(), default=1, nullable=False),
        sa.Column("callback_data", ExtendedJSON, nullable=False),
        sa.Column("callback_type", sa.String(20), nullable=False),
        sa.Column("dag_directory", sa.String(length=1000), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade():
    op.drop_table(TABLE_NAME)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Save Rendered Template Fields"""
from __future__ import annotations

import os
from typing import TYPE_CHECKING

import sqlalchemy_jsonfield
from sqlalchemy import Column, ForeignKeyConstraint, Integer, PrimaryKeyConstraint, text
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import Session, relationship

from airflow.configuration import conf
from airflow.models.base import Base, StringID
from airflow.models.taskinstance import TaskInstance
from airflow.serialization.helpers import serialize_template_field
from airflow.settings import json
from airflow.utils.retries import retry_db_transaction
from airflow.utils.session import NEW_SESSION, provide_session
from airflow.utils.sqlalchemy import tuple_not_in_condition

if TYPE_CHECKING:
    from sqlalchemy.sql import FromClause


class RenderedTaskInstanceFields(Base):
    """Save Rendered Template Fields"""

    __tablename__ = "rendered_task_instance_fields"

    dag_id = Column(StringID(), primary_key=True)
    task_id = Column(StringID(), primary_key=True)
    run_id = Column(StringID(), primary_key=True)
    map_index = Column(Integer, primary_key=True, server_default=text("-1"))
    rendered_fields = Column(sqlalchemy_jsonfield.JSONField(json=json), nullable=False)
    k8s_pod_yaml = Column(sqlalchemy_jsonfield.JSONField(json=json), nullable=True)

    __table_args__ = (
        PrimaryKeyConstraint(
            "dag_id",
            "task_id",
            "run_id",
            "map_index",
            name="rendered_task_instance_fields_pkey",
            mssql_clustered=True,
        ),
        ForeignKeyConstraint(
            [dag_id, task_id, run_id, map_index],
            [
                "task_instance.dag_id",
                "task_instance.task_id",
                "task_instance.run_id",
                "task_instance.map_index",
            ],
            name="rtif_ti_fkey",
            ondelete="CASCADE",
        ),
    )
    task_instance = relationship(
        "TaskInstance",
        lazy="joined",
        back_populates="rendered_task_instance_fields",
    )

    # We don't need a DB level FK here, as we already have that to TI (which has one to DR) but by defining
    # the relationship we can more easily find the execution date for these rows
    dag_run = relationship(
        "DagRun",
        primaryjoin="""and_(
            RenderedTaskInstanceFields.dag_id == foreign(DagRun.dag_id),
            RenderedTaskInstanceFields.run_id == foreign(DagRun.run_id),
        )""",
        viewonly=True,
    )

    execution_date = association_proxy("dag_run", "execution_date")

    def __init__(self, ti: TaskInstance, render_templates=True):
        self.dag_id = ti.dag_id
        self.task_id = ti.task_id
        self.run_id = ti.run_id
        self.map_index = ti.map_index
        self.ti = ti
        if render_templates:
            ti.render_templates()
        self.task = ti.task
        if os.environ.get("AIRFLOW_IS_K8S_EXECUTOR_POD", None):
            self.k8s_pod_yaml = ti.render_k8s_pod_yaml()
        self.rendered_fields = {
            field: serialize_template_field(getattr(self.task, field)) for field in self.task.template_fields
        }

        self._redact()

    def __repr__(self):
        prefix = f"<{self.__class__.__name__}: {self.dag_id}.{self.task_id} {self.run_id}"
        if self.map_index != -1:
            prefix += f" map_index={self.map_index}"
        return prefix + ">"

    def _redact(self):
        from airflow.utils.log.secrets_masker import redact

        if self.k8s_pod_yaml:
            self.k8s_pod_yaml = redact(self.k8s_pod_yaml)

        for field, rendered in self.rendered_fields.items():
            self.rendered_fields[field] = redact(rendered, field)

    @classmethod
    @provide_session
    def get_templated_fields(cls, ti: TaskInstance, session: Session = NEW_SESSION) -> dict | None:
        """
        Get templated field for a TaskInstance from the RenderedTaskInstanceFields
        table.

        :param ti: Task Instance
        :param session: SqlAlchemy Session
        :return: Rendered Templated TI field
        """
        result = (
            session.query(cls.rendered_fields)
            .filter(
                cls.dag_id == ti.dag_id,
                cls.task_id == ti.task_id,
                cls.run_id == ti.run_id,
                cls.map_index == ti.map_index,
            )
            .one_or_none()
        )

        if result:
            rendered_fields = result.rendered_fields
            return rendered_fields
        else:
            return None

    @classmethod
    @provide_session
    def get_k8s_pod_yaml(cls, ti: TaskInstance, session: Session = NEW_SESSION) -> dict | None:
        """
        Get rendered Kubernetes Pod Yaml for a TaskInstance from the RenderedTaskInstanceFields
        table.

        :param ti: Task Instance
        :param session: SqlAlchemy Session
        :return: Kubernetes Pod Yaml
        """
        result = (
            session.query(cls.k8s_pod_yaml)
            .filter(
                cls.dag_id == ti.dag_id,
                cls.task_id == ti.task_id,
                cls.run_id == ti.run_id,
                cls.map_index == ti.map_index,
            )
            .one_or_none()
        )
        return result.k8s_pod_yaml if result else None

    @provide_session
    def write(self, session: Session = None):
        """Write instance to database

        :param session: SqlAlchemy Session
        """
        session.merge(self)

    @classmethod
    @provide_session
    def delete_old_records(
        cls,
        task_id: str,
        dag_id: str,
        num_to_keep: int = conf.getint("core", "max_num_rendered_ti_fields_per_task", fallback=0),
        session: Session = NEW_SESSION,
    ) -> None:
        """
        Keep only Last X (num_to_keep) number of records for a task by deleting others.

        In the case of data for a mapped task either all of the rows or none of the rows will be deleted, so
        we don't end up with partial data for a set of mapped Task Instances left in the database.

        :param task_id: Task ID
        :param dag_id: Dag ID
        :param num_to_keep: Number of Records to keep
        :param session: SqlAlchemy Session
        """
        from airflow.models.dagrun import DagRun

        if num_to_keep <= 0:
            return

        tis_to_keep_query = (
            session.query(cls.dag_id, cls.task_id, cls.run_id)
            .filter(cls.dag_id == dag_id, cls.task_id == task_id)
            .join(cls.dag_run)
            .distinct()
            .order_by(DagRun.execution_date.desc())
            .limit(num_to_keep)
        )

        cls._do_delete_old_records(
            dag_id=dag_id,
            task_id=task_id,
            ti_clause=tis_to_keep_query.subquery(),
            session=session,
        )
        session.flush()

    @classmethod
    @retry_db_transaction
    def _do_delete_old_records(
        cls,
        *,
        task_id: str,
        dag_id: str,
        ti_clause: FromClause,
        session: Session,
    ) -> None:
        # This query might deadlock occasionally and it should be retried if fails (see decorator)
        session.query(cls).filter(
            cls.dag_id == dag_id,
            cls.task_id == task_id,
            tuple_not_in_condition(
                (cls.dag_id, cls.task_id, cls.run_id),
                session.query(ti_clause.c.dag_id, ti_clause.c.task_id, ti_clause.c.run_id),
            ),
        ).delete(synchronize_session=False)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime

# This product contains a modified portion of 'Flask App Builder' developed by Daniel Vaz Gaspar.
# (https://github.com/dpgaspar/Flask-AppBuilder).
# Copyright 2013, Daniel Vaz Gaspar
from typing import TYPE_CHECKING

from flask import current_app, g
from flask_appbuilder.models.sqla import Model
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    String,
    Table,
    UniqueConstraint,
    event,
    func,
)
from sqlalchemy.orm import backref, declared_attr, relationship

from airflow.models.base import Base

"""
Compatibility note: The models in this file are duplicated from Flask AppBuilder.
"""
# Use airflow metadata to create the tables
Model.metadata = Base.metadata

if TYPE_CHECKING:
    try:
        from sqlalchemy import Identity
    except Exception:
        Identity = None


class Action(Model):
    """Represents permission actions such as `can_read`."""

    __tablename__ = "ab_permission"
    id = Column(Integer, primary_key=True)
    name = Column(String(100), unique=True, nullable=False)

    def __repr__(self):
        return self.name


class Resource(Model):
    """Represents permission object such as `User` or `Dag`."""

    __tablename__ = "ab_view_menu"
    id = Column(Integer, primary_key=True)
    name = Column(String(250), unique=True, nullable=False)

    def __eq__(self, other):
        return (isinstance(other, self.__class__)) and (self.name == other.name)

    def __neq__(self, other):
        return self.name != other.name

    def __repr__(self):
        return self.name


assoc_permission_role = Table(
    "ab_permission_view_role",
    Model.metadata,
    Column("id", Integer, primary_key=True),
    Column("permission_view_id", Integer, ForeignKey("ab_permission_view.id")),
    Column("role_id", Integer, ForeignKey("ab_role.id")),
    UniqueConstraint("permission_view_id", "role_id"),
)


class Role(Model):
    """Represents a user role to which permissions can be assigned."""

    __tablename__ = "ab_role"

    id = Column(Integer, primary_key=True)
    name = Column(String(64), unique=True, nullable=False)
    permissions = relationship("Permission", secondary=assoc_permission_role, backref="role", lazy="joined")

    def __repr__(self):
        return self.name


class Permission(Model):
    """Permission pair comprised of an Action + Resource combo."""

    __tablename__ = "ab_permission_view"
    __table_args__ = (UniqueConstraint("permission_id", "view_menu_id"),)
    id = Column(Integer, primary_key=True)
    action_id = Column("permission_id", Integer, ForeignKey("ab_permission.id"))
    action = relationship(
        "Action",
        uselist=False,
        lazy="joined",
    )
    resource_id = Column("view_menu_id", Integer, ForeignKey("ab_view_menu.id"))
    resource = relationship(
        "Resource",
        uselist=False,
        lazy="joined",
    )

    def __repr__(self):
        return str(self.action).replace("_", " ") + " on " + str(self.resource)


assoc_user_role = Table(
    "ab_user_role",
    Model.metadata,
    Column("id", Integer, primary_key=True),
    Column("user_id", Integer, ForeignKey("ab_user.id")),
    Column("role_id", Integer, ForeignKey("ab_role.id")),
    UniqueConstraint("user_id", "role_id"),
)


class User(Model):
    """Represents an Airflow user which has roles assigned to it."""

    __tablename__ = "ab_user"
    id = Column(Integer, primary_key=True)
    first_name = Column(String(256), nullable=False)
    last_name = Column(String(256), nullable=False)
    username = Column(
        String(512).with_variant(String(512, collation="NOCASE"), "sqlite"), unique=True, nullable=False
    )
    password = Column(String(256))
    active = Column(Boolean)
    email = Column(String(512), unique=True, nullable=False)
    last_login = Column(DateTime)
    login_count = Column(Integer)
    fail_login_count = Column(Integer)
    roles = relationship("Role", secondary=assoc_user_role, backref="user", lazy="selectin")
    created_on = Column(DateTime, default=datetime.datetime.now, nullable=True)
    changed_on = Column(DateTime, default=datetime.datetime.now, nullable=True)

    @declared_attr
    def created_by_fk(self):
        return Column(Integer, ForeignKey("ab_user.id"), default=self.get_user_id, nullable=True)

    @declared_attr
    def changed_by_fk(self):
        return Column(Integer, ForeignKey("ab_user.id"), default=self.get_user_id, nullable=True)

    created_by = relationship(
        "User",
        backref=backref("created", uselist=True),
        remote_side=[id],
        primaryjoin="User.created_by_fk == User.id",
        uselist=False,
    )
    changed_by = relationship(
        "User",
        backref=backref("changed", uselist=True),
        remote_side=[id],
        primaryjoin="User.changed_by_fk == User.id",
        uselist=False,
    )

    @classmethod
    def get_user_id(cls):
        try:
            return g.user.id
        except Exception:
            return None

    @property
    def is_authenticated(self):
        return True

    @property
    def is_active(self):
        return self.active

    @property
    def is_anonymous(self):
        return False

    @property
    def perms(self):
        if not self._perms:
            # Using the ORM here is _slow_ (Creating lots of objects to then throw them away) since this is in
            # the path for every request. Avoid it if we can!
            if current_app:
                sm = current_app.appbuilder.sm
                self._perms: set[tuple[str, str]] = set(
                    sm.get_session.query(sm.action_model.name, sm.resource_model.name)
                    .join(sm.permission_model.action)
                    .join(sm.permission_model.resource)
                    .join(sm.permission_model.role)
                    .filter(sm.role_model.user.contains(self))
                    .all()
                )
            else:
                self._perms = {
                    (perm.action.name, perm.resource.name) for role in self.roles for perm in role.permissions
                }
        return self._perms

    def get_id(self):
        return self.id

    def get_full_name(self):
        return f"{self.first_name} {self.last_name}"

    def __repr__(self):
        return self.get_full_name()

    _perms = None


class RegisterUser(Model):
    """Represents a user registration."""

    __tablename__ = "ab_register_user"
    id = Column(Integer, primary_key=True)
    first_name = Column(String(256), nullable=False)
    last_name = Column(String(256), nullable=False)
    username = Column(
        String(512).with_variant(String(512, collation="NOCASE"), "sqlite"), unique=True, nullable=False
    )
    password = Column(String(256))
    email = Column(String(512), nullable=False)
    registration_date = Column(DateTime, default=datetime.datetime.now, nullable=True)
    registration_hash = Column(String(256))


@event.listens_for(User.__table__, "before_create")
def add_index_on_ab_user_username_postgres(table, conn, **kw):
    if conn.dialect.name != "postgresql":
        return
    table.indexes.add(Index("idx_ab_user_username", func.lower(table.c.username), unique=True))


@event.listens_for(RegisterUser.__table__, "before_create")
def add_index_on_ab_register_user_username_postgres(table, conn, **kw):
    if conn.dialect.name != "postgresql":
        return
    table.indexes.add(Index("idx_ab_register_user_username", func.lower(table.c.username), unique=True))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from datetime import datetime
from typing import Optional

from pydantic import BaseModel as BaseModelPydantic

from airflow.jobs.base_job_runner import BaseJobRunner


def check_runner_initialized(job_runner: Optional[BaseJobRunner], job_type: str) -> BaseJobRunner:
    if job_runner is None:
        raise ValueError(f"In order to run {job_type} you need to initialize the {job_type}Runner first.")
    return job_runner


class JobPydantic(BaseModelPydantic):
    """Serializable representation of the Job ORM SqlAlchemyModel used by internal API"""

    id: Optional[int]
    dag_id: Optional[str]
    state: Optional[str]
    job_type: Optional[str]
    start_date: Optional[datetime]
    end_date: Optional[datetime]
    latest_heartbeat: datetime
    executor_class: Optional[str]
    hostname: Optional[str]
    unixname: Optional[str]

    # not an ORM field
    heartrate: Optional[int]
    max_tis_per_query: Optional[int]

    class Config:
        """Make sure it deals automatically with ORM classes of SQL Alchemy"""

        orm_mode = True



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow.api_connexion.schemas.health_schema import health_schema
from airflow.api_connexion.types import APIResponse
from airflow.jobs.scheduler_job_runner import SchedulerJobRunner
from airflow.jobs.triggerer_job_runner import TriggererJobRunner

HEALTHY = "healthy"
UNHEALTHY = "unhealthy"


def get_health() -> APIResponse:
    """Return the health of the airflow scheduler and metadatabase."""
    metadatabase_status = HEALTHY
    latest_scheduler_heartbeat = None
    latest_triggerer_heartbeat = None
    scheduler_status = UNHEALTHY
    triggerer_status: str | None = UNHEALTHY
    try:
        scheduler_job = SchedulerJobRunner.most_recent_job()

        if scheduler_job:
            latest_scheduler_heartbeat = scheduler_job.latest_heartbeat.isoformat()
            if scheduler_job.is_alive():
                scheduler_status = HEALTHY
    except Exception:
        metadatabase_status = UNHEALTHY
    try:
        triggerer_job = TriggererJobRunner.most_recent_job()

        if triggerer_job:
            latest_triggerer_heartbeat = triggerer_job.latest_heartbeat.isoformat()
            if triggerer_job.is_alive():
                triggerer_status = HEALTHY
        else:
            triggerer_status = None
    except Exception:
        metadatabase_status = UNHEALTHY

    payload = {
        "metadatabase": {"status": metadatabase_status},
        "scheduler": {
            "status": scheduler_status,
            "latest_scheduler_heartbeat": latest_scheduler_heartbeat,
        },
        "triggerer": {
            "status": triggerer_status,
            "latest_triggerer_heartbeat": latest_triggerer_heartbeat,
        },
    }

    return health_schema.dump(payload)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Platform and system specific function."""
from __future__ import annotations

import getpass
import logging
import os
import pkgutil
import platform
import sys

from airflow.compat.functools import cache

IS_WINDOWS = platform.system() == "Windows"

log = logging.getLogger(__name__)


def is_tty():
    """
    Checks if the standard output is connected (is associated with a terminal device) to a tty(-like)
    device.
    """
    if not hasattr(sys.stdout, "isatty"):
        return False
    return sys.stdout.isatty()


def is_terminal_support_colors() -> bool:
    """Try to determine if the current terminal supports colors."""
    if sys.platform == "win32":
        return False
    if not is_tty():
        return False
    if "COLORTERM" in os.environ:
        return True
    term = os.environ.get("TERM", "dumb").lower()
    if term in ("xterm", "linux") or "color" in term:
        return True
    return False


def get_airflow_git_version():
    """Returns the git commit hash representing the current version of the application."""
    git_version = None
    try:
        git_version = str(pkgutil.get_data("airflow", "git_version"), encoding="UTF-8")
    except Exception as e:
        log.debug(e)

    return git_version


@cache
def getuser() -> str:
    """
    Gets the username associated with the current user, or error with a nice
    error message if there's no current user.

    We don't want to fall back to os.getuid() because not having a username
    probably means the rest of the user environment is wrong (e.g. no $HOME).
    Explicit failure is better than silently trying to work badly.
    """
    try:
        return getpass.getuser()
    except KeyError:
        # Inner import to avoid circular import
        from airflow.exceptions import AirflowConfigException

        raise AirflowConfigException(
            "The user that Airflow is running as has no username; you must run"
            "Airflow as a full user, with a username and home directory, "
            "in order for it to function properly."
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import collections
import collections.abc
import functools
from typing import TYPE_CHECKING, Iterator, NamedTuple

from sqlalchemy import and_, func, or_

from airflow.models import MappedOperator
from airflow.models.taskinstance import PAST_DEPENDS_MET
from airflow.ti_deps.dep_context import DepContext
from airflow.ti_deps.deps.base_ti_dep import BaseTIDep, TIDepStatus
from airflow.utils.state import TaskInstanceState
from airflow.utils.trigger_rule import TriggerRule as TR

if TYPE_CHECKING:
    from sqlalchemy.orm import Session
    from sqlalchemy.sql.expression import ColumnOperators

    from airflow.models.taskinstance import TaskInstance


class _UpstreamTIStates(NamedTuple):
    """States of the upstream tis for a specific ti.

    This is used to determine whether the specific ti can run in this iteration.
    """

    success: int
    skipped: int
    failed: int
    upstream_failed: int
    removed: int
    done: int
    success_setup: int
    skipped_setup: int

    @classmethod
    def calculate(cls, finished_upstreams: Iterator[TaskInstance]) -> _UpstreamTIStates:
        """Calculate states for a task instance.

        ``counter`` is inclusive of ``setup_counter`` -- e.g. if there are 2 skipped upstreams, one
        of which is a setup, then counter will show 2 skipped and setup counter will show 1.

        :param ti: the ti that we want to calculate deps for
        :param finished_tis: all the finished tasks of the dag_run
        """
        counter: dict[str, int] = collections.Counter()
        setup_counter: dict[str, int] = collections.Counter()
        for ti in finished_upstreams:
            curr_state = {ti.state: 1}
            counter.update(curr_state)
            # setup task cannot be mapped
            if not isinstance(ti.task, MappedOperator) and ti.task._is_setup:
                setup_counter.update(curr_state)
        return _UpstreamTIStates(
            success=counter.get(TaskInstanceState.SUCCESS, 0),
            skipped=counter.get(TaskInstanceState.SKIPPED, 0),
            failed=counter.get(TaskInstanceState.FAILED, 0),
            upstream_failed=counter.get(TaskInstanceState.UPSTREAM_FAILED, 0),
            removed=counter.get(TaskInstanceState.REMOVED, 0),
            done=sum(counter.values()),
            success_setup=setup_counter.get(TaskInstanceState.SUCCESS, 0),
            skipped_setup=setup_counter.get(TaskInstanceState.SKIPPED, 0),
        )


class TriggerRuleDep(BaseTIDep):
    """
    Determines if a task's upstream tasks are in a state that allows a given task instance
    to run.
    """

    NAME = "Trigger Rule"
    IGNORABLE = True
    IS_TASK_DEP = True

    def _get_dep_statuses(
        self,
        ti: TaskInstance,
        session: Session,
        dep_context: DepContext,
    ) -> Iterator[TIDepStatus]:
        # Checking that all upstream dependencies have succeeded.
        if not ti.task.upstream_task_ids:
            yield self._passing_status(reason="The task instance did not have any upstream tasks.")
            return
        if ti.task.trigger_rule == TR.ALWAYS:
            yield self._passing_status(reason="The task had a always trigger rule set.")
            return
        yield from self._evaluate_trigger_rule(ti=ti, dep_context=dep_context, session=session)

    def _evaluate_trigger_rule(
        self,
        *,
        ti: TaskInstance,
        dep_context: DepContext,
        session: Session,
    ) -> Iterator[TIDepStatus]:
        """Evaluate whether ``ti``'s trigger rule was met.

        :param ti: Task instance to evaluate the trigger rule of.
        :param dep_context: The current dependency context.
        :param session: Database session.
        """
        from airflow.models.abstractoperator import NotMapped
        from airflow.models.expandinput import NotFullyPopulated
        from airflow.models.operator import needs_expansion
        from airflow.models.taskinstance import TaskInstance

        task = ti.task
        upstream_tasks = {t.task_id: t for t in task.upstream_list}
        trigger_rule = task.trigger_rule

        @functools.lru_cache()
        def _get_expanded_ti_count() -> int:
            """Get how many tis the current task is supposed to be expanded into.

            This extra closure allows us to query the database only when needed,
            and at most once.
            """
            return task.get_mapped_ti_count(ti.run_id, session=session)

        @functools.lru_cache()
        def _get_relevant_upstream_map_indexes(upstream_id: str) -> int | range | None:
            """Get the given task's map indexes relevant to the current ti.

            This extra closure allows us to query the database only when needed,
            and at most once for each task (instead of once for each expanded
            task instance of the same task).
            """
            try:
                expanded_ti_count = _get_expanded_ti_count()
            except (NotFullyPopulated, NotMapped):
                return None
            return ti.get_relevant_upstream_map_indexes(
                upstream_tasks[upstream_id],
                expanded_ti_count,
                session=session,
            )

        def _is_relevant_upstream(upstream: TaskInstance) -> bool:
            """Whether a task instance is a "relevant upstream" of the current task."""
            # Not actually an upstream task.
            if upstream.task_id not in task.upstream_task_ids:
                return False
            # The current task is not in a mapped task group. All tis from an
            # upstream task are relevant.
            if task.get_closest_mapped_task_group() is None:
                return True
            # The upstream ti is not expanded. The upstream may be mapped or
            # not, but the ti is relevant either way.
            if upstream.map_index < 0:
                return True
            # Now we need to perform fine-grained check on whether this specific
            # upstream ti's map index is relevant.
            relevant = _get_relevant_upstream_map_indexes(upstream.task_id)
            if relevant is None:
                return True
            if relevant == upstream.map_index:
                return True
            if isinstance(relevant, collections.abc.Container) and upstream.map_index in relevant:
                return True
            return False

        finished_upstream_tis = (
            finished_ti
            for finished_ti in dep_context.ensure_finished_tis(ti.get_dagrun(session), session)
            if _is_relevant_upstream(finished_ti)
        )
        upstream_states = _UpstreamTIStates.calculate(finished_upstream_tis)

        success = upstream_states.success
        skipped = upstream_states.skipped
        failed = upstream_states.failed
        upstream_failed = upstream_states.upstream_failed
        removed = upstream_states.removed
        done = upstream_states.done
        success_setup = upstream_states.success_setup
        skipped_setup = upstream_states.skipped_setup

        def _iter_upstream_conditions() -> Iterator[ColumnOperators]:
            # Optimization: If the current task is not in a mapped task group,
            # it depends on all upstream task instances.
            if task.get_closest_mapped_task_group() is None:
                yield TaskInstance.task_id.in_(upstream_tasks)
                return
            # Otherwise we need to figure out which map indexes are depended on
            # for each upstream by the current task instance.
            for upstream_id in upstream_tasks:
                map_indexes = _get_relevant_upstream_map_indexes(upstream_id)
                if map_indexes is None:  # All tis of this upstream are dependencies.
                    yield (TaskInstance.task_id == upstream_id)
                    continue
                # At this point we know we want to depend on only selected tis
                # of this upstream task. Since the upstream may not have been
                # expanded at this point, we also depend on the non-expanded ti
                # to ensure at least one ti is included for the task.
                yield and_(TaskInstance.task_id == upstream_id, TaskInstance.map_index < 0)
                if isinstance(map_indexes, range) and map_indexes.step == 1:
                    yield and_(
                        TaskInstance.task_id == upstream_id,
                        TaskInstance.map_index >= map_indexes.start,
                        TaskInstance.map_index < map_indexes.stop,
                    )
                elif isinstance(map_indexes, collections.abc.Container):
                    yield and_(TaskInstance.task_id == upstream_id, TaskInstance.map_index.in_(map_indexes))
                else:
                    yield and_(TaskInstance.task_id == upstream_id, TaskInstance.map_index == map_indexes)

        # Optimization: Don't need to hit the database if all upstreams are
        # "simple" tasks (no task or task group mapping involved).
        if not any(needs_expansion(t) for t in upstream_tasks.values()):
            upstream = len(upstream_tasks)
            upstream_setup = len(
                [x for x in upstream_tasks.values() if not isinstance(x, MappedOperator) and x._is_setup]
            )
        else:
            upstream = (
                session.query(func.count())
                .filter(TaskInstance.dag_id == ti.dag_id, TaskInstance.run_id == ti.run_id)
                .filter(or_(*_iter_upstream_conditions()))
                .scalar()
            )
            # todo: add support for mapped setup?
            upstream_setup = None
        upstream_done = done >= upstream

        changed = False
        new_state = None
        if dep_context.flag_upstream_failed:
            if trigger_rule == TR.ALL_SUCCESS:
                if upstream_failed or failed:
                    new_state = TaskInstanceState.UPSTREAM_FAILED
                elif skipped:
                    new_state = TaskInstanceState.SKIPPED
                elif removed and success and ti.map_index > -1:
                    if ti.map_index >= success:
                        new_state = TaskInstanceState.REMOVED
            elif trigger_rule == TR.ALL_FAILED:
                if success or skipped:
                    new_state = TaskInstanceState.SKIPPED
            elif trigger_rule == TR.ONE_SUCCESS:
                if upstream_done and done == skipped:
                    # if upstream is done and all are skipped mark as skipped
                    new_state = TaskInstanceState.SKIPPED
                elif upstream_done and success <= 0:
                    # if upstream is done and there are no success mark as upstream failed
                    new_state = TaskInstanceState.UPSTREAM_FAILED
            elif trigger_rule == TR.ONE_FAILED:
                if upstream_done and not (failed or upstream_failed):
                    new_state = TaskInstanceState.SKIPPED
            elif trigger_rule == TR.ONE_DONE:
                if upstream_done and not (failed or success):
                    new_state = TaskInstanceState.SKIPPED
            elif trigger_rule == TR.NONE_FAILED:
                if upstream_failed or failed:
                    new_state = TaskInstanceState.UPSTREAM_FAILED
            elif trigger_rule == TR.NONE_FAILED_MIN_ONE_SUCCESS:
                if upstream_failed or failed:
                    new_state = TaskInstanceState.UPSTREAM_FAILED
                elif skipped == upstream:
                    new_state = TaskInstanceState.SKIPPED
            elif trigger_rule == TR.NONE_SKIPPED:
                if skipped:
                    new_state = TaskInstanceState.SKIPPED
            elif trigger_rule == TR.ALL_SKIPPED:
                if success or failed:
                    new_state = TaskInstanceState.SKIPPED
            elif trigger_rule == TR.ALL_DONE_SETUP_SUCCESS:
                if upstream_done and upstream_setup and skipped_setup >= upstream_setup:
                    # when there is an upstream setup and they have all skipped, then skip
                    new_state = TaskInstanceState.SKIPPED
                elif upstream_done and upstream_setup and success_setup == 0:
                    # when there is an upstream setup, if none succeeded, mark upstream failed
                    # if at least one setup ran, we'll let it run
                    new_state = TaskInstanceState.UPSTREAM_FAILED
        if new_state is not None:
            if new_state == TaskInstanceState.SKIPPED and dep_context.wait_for_past_depends_before_skipping:
                past_depends_met = ti.xcom_pull(
                    task_ids=ti.task_id, key=PAST_DEPENDS_MET, session=session, default=False
                )
                if not past_depends_met:
                    yield self._failing_status(
                        reason=("Task should be skipped but the the past depends are not met")
                    )
                    return
            changed = ti.set_state(new_state, session)

        if changed:
            dep_context.have_changed_ti_states = True

        if trigger_rule == TR.ONE_SUCCESS:
            if success <= 0:
                yield self._failing_status(
                    reason=(
                        f"Task's trigger rule '{trigger_rule}' requires one upstream task success, "
                        f"but none were found. upstream_states={upstream_states}, "
                        f"upstream_task_ids={task.upstream_task_ids}"
                    )
                )
        elif trigger_rule == TR.ONE_FAILED:
            if not failed and not upstream_failed:
                yield self._failing_status(
                    reason=(
                        f"Task's trigger rule '{trigger_rule}' requires one upstream task failure, "
                        f"but none were found. upstream_states={upstream_states}, "
                        f"upstream_task_ids={task.upstream_task_ids}"
                    )
                )
        elif trigger_rule == TR.ONE_DONE:
            if success + failed <= 0:
                yield self._failing_status(
                    reason=(
                        f"Task's trigger rule '{trigger_rule}'"
                        "requires at least one upstream task failure or success"
                        f"but none were failed or success. upstream_states={upstream_states}, "
                        f"upstream_task_ids={task.upstream_task_ids}"
                    )
                )
        elif trigger_rule == TR.ALL_SUCCESS:
            num_failures = upstream - success
            if ti.map_index > -1:
                num_failures -= removed
            if num_failures > 0:
                yield self._failing_status(
                    reason=(
                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to have "
                        f"succeeded, but found {num_failures} non-success(es). "
                        f"upstream_states={upstream_states}, "
                        f"upstream_task_ids={task.upstream_task_ids}"
                    )
                )
        elif trigger_rule == TR.ALL_FAILED:
            num_success = upstream - failed - upstream_failed
            if ti.map_index > -1:
                num_success -= removed
            if num_success > 0:
                yield self._failing_status(
                    reason=(
                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to have failed, "
                        f"but found {num_success} non-failure(s). "
                        f"upstream_states={upstream_states}, "
                        f"upstream_task_ids={task.upstream_task_ids}"
                    )
                )
        elif trigger_rule == TR.ALL_DONE:
            if not upstream_done:
                yield self._failing_status(
                    reason=(
                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to have "
                        f"completed, but found {len(upstream_tasks) - done} task(s) that were not done. "
                        f"upstream_states={upstream_states}, "
                        f"upstream_task_ids={task.upstream_task_ids}"
                    )
                )
        elif trigger_rule == TR.NONE_FAILED:
            num_failures = upstream - success - skipped
            if ti.map_index > -1:
                num_failures -= removed
            if num_failures > 0:
                yield self._failing_status(
                    reason=(
                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to have "
                        f"succeeded or been skipped, but found {num_failures} non-success(es). "
                        f"upstream_states={upstream_states}, "
                        f"upstream_task_ids={task.upstream_task_ids}"
                    )
                )
        elif trigger_rule == TR.NONE_FAILED_MIN_ONE_SUCCESS:
            num_failures = upstream - success - skipped
            if ti.map_index > -1:
                num_failures -= removed
            if num_failures > 0:
                yield self._failing_status(
                    reason=(
                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to have "
                        f"succeeded or been skipped, but found {num_failures} non-success(es). "
                        f"upstream_states={upstream_states}, "
                        f"upstream_task_ids={task.upstream_task_ids}"
                    )
                )
        elif trigger_rule == TR.NONE_SKIPPED:
            if not upstream_done or (skipped > 0):
                yield self._failing_status(
                    reason=(
                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to not have been "
                        f"skipped, but found {skipped} task(s) skipped. "
                        f"upstream_states={upstream_states}, "
                        f"upstream_task_ids={task.upstream_task_ids}"
                    )
                )
        elif trigger_rule == TR.ALL_SKIPPED:
            num_non_skipped = upstream - skipped
            if num_non_skipped > 0:
                yield self._failing_status(
                    reason=(
                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to have been "
                        f"skipped, but found {num_non_skipped} task(s) in non skipped state. "
                        f"upstream_states={upstream_states}, "
                        f"upstream_task_ids={task.upstream_task_ids}"
                    )
                )
        elif trigger_rule == TR.ALL_DONE_SETUP_SUCCESS:
            if not upstream_done:
                yield self._failing_status(
                    reason=(
                        f"Task's trigger rule '{trigger_rule}' requires all upstream tasks to have "
                        f"completed, but found {len(upstream_tasks) - done} task(s) that were not done. "
                        f"upstream_states={upstream_states}, "
                        f"upstream_task_ids={task.upstream_task_ids}"
                    )
                )
            elif upstream_setup is None:  # for now, None only happens in mapped case
                yield self._failing_status(
                    reason=(
                        f"Task's trigger rule '{trigger_rule}' cannot have mapped tasks as upstream. "
                        f"upstream_states={upstream_states}, "
                        f"upstream_task_ids={task.upstream_task_ids}"
                    )
                )
            elif upstream_setup and not success_setup >= 1:
                yield self._failing_status(
                    reason=(
                        f"Task's trigger rule '{trigger_rule}' requires at least one upstream setup task be "
                        f"successful, but found {upstream_setup - success_setup} task(s) that were not. "
                        f"upstream_states={upstream_states}, "
                        f"upstream_task_ids={task.upstream_task_ids}"
                    )
                )
        else:
            yield self._failing_status(reason=f"No strategy to evaluate trigger rule '{trigger_rule}'.")
